#!/usr/bin/env python

#############################################################################
##
# This file is part of Taurus
##
# http://taurus-scada.org
##
# Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
# Taurus is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
##
# Taurus is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
##
# You should have received a copy of the GNU Lesser General Public License
# along with Taurus.  If not, see <http://www.gnu.org/licenses/>.
##
#############################################################################

"""
taurusgrid.py: module containing the Taurus Widget: TaurusGrid
original idea and development by gcuni
integrated with taurus and regular expressions by srubio
alba, 2009
"""

# This module needs a total cleanup. Both re. code conventions and algorithms.
#   --cpascual 20140827

from __future__ import print_function
from future.utils import string_types
from future import standard_library
standard_library.install_aliases()
from builtins import str


__all__ = ["TaurusGrid"]

__docformat__ = 'restructuredtext'

import re
from queue import Queue

from taurus.external.qt import Qt, QtGui, QtCore

import taurus
from taurus.qt.qtcore.util.emitter import (modelSetter,
                                           SingletonWorker, MethodModel)
from taurus.core.taurusmanager import TaurusManager
from taurus.qt.qtgui.base import TaurusBaseWidget
from taurus.qt.qtgui.panel import TaurusValue

metachars = re.compile(r'([.][*])|([.][^*])|([$^+\-?{}\[\]|()])')


def re_search_low(regexp, target):
    return re.search(regexp.lower(), target.lower())


def re_match_low(regexp, target):
    return re.match(regexp.lower(), target.lower())


def get_all_models(expressions, limit=1000):
    '''
    All devices matching expressions must be obtained.
    For each device only the good attributes are read.

    It practically equals to fandango.get_matching_attributes; check which is better!
    Move this method to taurus.core.tango.search
    '''
    # print( 'In TaurusGrid.get_all_models(%s:"%s") ...' % (type(expressions),expressions))
    if isinstance(expressions, string_types):
        # if any(re.match(s,expressions) for s in ('\{.*\}','\(.*\)','\[.*\]')):
        ##self.debug( 'evaluating expressions ....')
        # expressions = list(eval(expressions))
        # else:
        ##self.debug( 'expressions as string separated by commas ...')
        expressions = expressions.split(',')

    elif isinstance(expressions, (list, tuple, dict)):
        # self.debug( 'expressions converted from list ...')
        expressions = list(str(e) for e in expressions)

    # self.debug( 'In TaurusGrid.get_all_models(%s:"%s") ...' % (type(expressions),expressions))
    taurus_db = taurus.Authority()
    # taurus_db = taurus.Authority(os.environ['TANGO_HOST'])
    # WHAAAAAAT????? Someone should get beaten for this line
    if 'SimulationAuthority' in str(type(taurus_db)):
        # self.trace( 'Using a simulated database ...')
        models = expressions
    else:
        all_devs = taurus_db.get_device_exported('*')
        models = []
        for exp in expressions:
            # self.trace( 'evaluating exp = "%s"' % exp)
            exp = str(exp)
            devs = []
            targets = []
            if exp.count('/') == 3:
                device, attribute = exp.rsplit('/', 1)
            else:
                device, attribute = exp, 'State'

            if any(c in device for c in '.*[]()+?'):
                if '*' in device and '.*' not in device:
                    device = device.replace('*', '.*')
                devs = [s for s in all_devs if re_match_low(device, s)]
            else:
                devs = [device]

            # self.trace( 'TaurusGrid.get_all_models(): devices matched by %s / %s are %d:' % (device,attribute,len(devs)))
            # self.debug( '%s' % (devs))
            for dev in devs:
                if any(c in attribute for c in '.*[]()+?'):
                    if '*' in attribute and '.*' not in attribute:
                        attribute = attribute.replace('*', '.*')
                    try:
                        # taurus_dp = taurus.core.taurusdevice.TaurusDevice(dev)
                        taurus_dp = taurus.core.taurusmanager.TaurusManager().getFactory()().getDevice(
                            dev)
                        # self.debug( "taurus_dp = %s"%taurus_dp.getFullName())
                        attrs = [att.name for att in
                                 taurus_dp.attribute_list_query(
                                 ) if re_match_low(attribute, att.name)]
                        targets.extend(dev + '/' + att for att in attrs)
                    except Exception as e:
                        # self.warning( 'ERROR! TaurusGrid.get_all_models(): Unable to get attributes for device %s: %s' % (dev,str(e)))
                        pass
                else:
                    targets.append(dev + '/' + attribute)
            # print 'TaurusGrid.get_all_models(): targets added by %s are: %s'
            # % (exp,targets)
            models.extend(targets)
    models = models[:limit]
    # print( 'Out of TaurusGrid.get_all_models(...)')
    return models


def get_readwrite_models(expressions, limit=1000):
    '''
    All devices matching expressions must be obtained.
    For each device only the good attributes are read.
    '''
    # self.debug( 'In TaurusGrid.get_all_models(%s:"%s") ...' % (type(expressions),expressions))
    if isinstance(expressions, string_types):
        if any(re.match(s, expressions) for s in
               (r'\{.*\}', r'\(.*\)', r'\[.*\]')):
            # self.trace( 'evaluating expressions ....')
            expressions = list(eval(expressions))
        else:
            # self.trace( 'expressions as string separated by commas ...')
            expressions = expressions.split(',')

    elif isinstance(expressions, (list, tuple, dict)):
        expressions = list(str(e) for e in expressions)

    taurus_db = taurus.Authority()
    # WHAAAT???? At least check instances...
    if 'SimulationAuthority' in str(type(taurus_db)):
        models = expressions
    else:
        all_devs = taurus_db.get_device_exported('*')
        models = []
        for exp in expressions:
            exp = str(exp)
            devs = []
            targets = []
            if exp.count('/') == 3:
                device, attribute = exp.rsplit('/', 1)
            else:
                device, attribute = exp, 'State'

            if any(c in device for c in '.*[]()+?'):
                if '*' in device and '.*' not in device:
                    device = device.replace('*', '.*')
                devs = [s for s in all_devs if re_match_low(device, s)]
            else:
                devs = [device]

            for dev in devs:
                if any(c in attribute for c in '.*[]()+?'):
                    if '*' in attribute and '.*' not in attribute:
                        attribute = attribute.replace('*', '.*')
                    try:
                        taurus_dp = taurus.core.taurusmanager.TaurusManager().getFactory()().getDevice(
                            dev)
                        attrs = [att.name for att in
                                 taurus_dp.attribute_list_query(
                                 ) if re_match_low(attribute,
                                                   att.name) and att.isReadOnly()]
                        targets.extend(dev + '/' + att for att in attrs)
                    except Exception as e:
                        pass
                else:
                    targets.append(dev + '/' + attribute)
            models.extend(targets)
    models = models[:limit]
    return models


class TaurusGrid(QtGui.QFrame, TaurusBaseWidget):
    """ TaurusGrid is a Taurus widget designed to represent a set of attributes distributed in columns and rows.
    The Model will be a list with attributes or device names (for devices the State attribute will be shown).
    Each setModel(*) execution will be able to modify the attribute list.
    An example of execution:<pre>
    /usr/bin/python taurusgrid.py "model=lt.*/VC.*/.*/((C*)|(P*)|(I*))" cols=IP,CCG,PNV rows=LT01,LT02
    </pre>
    @author originally developed by gcuni, extended by srubio and sblanch
    @todo Future releases should allow a list of filters as argument
    @todo names/widgets should be accessible as a caselessdict dictionary (e.g. for adding custom context menus)
    @todo refactoring to have methods that add/remove new widgets one by one, not only the whole dictionary
    @todo _TAGS property should allow to change row/columns meaning and also add new Custom tags based on regexp
    """

    # -------------------------------------------------------------------------
    # Write your own code here to define the signals generated by this widget
    #

    itemSelected = Qt.pyqtSignal('QString')
    itemClicked = Qt.pyqtSignal('QString')

    _TAGS = ['DOMAIN', 'FAMILY', 'HOST',
             'LEVEL', 'CLASS', 'ATTRIBUTE', 'DEVICE']

    class _TaurusGridCell(Qt.QFrame):

        itemClicked = Qt.pyqtSignal('QString')

        # Done in this way as TaurusValue.mousePressEvent is never called
        def mousePressEvent(self, event):
            # print 'In cell clicked'
            targets = set(str(child.getModelName())
                          for child in self.children()
                          if hasattr(child, 'underMouse')
                          and child.underMouse()
                          and hasattr(child, 'getModelName')
                          )
            for t in targets:
                self.itemClicked.emit(t)

    def __init__(self, parent=None, designMode=False):
        name = self.__class__.__name__

        self.call__init__wo_kw(QtGui.QFrame, parent)
        # self.call__init__(TaurusBaseWidget, name, parent,
        #                   designMode=designMode)
        # It was needed to avoid exceptions in TaurusDesigner!
        if isinstance(parent, TaurusBaseWidget):
            self.call__init__(TaurusBaseWidget, name,
                              parent, designMode=designMode)
        else:
            self.call__init__(TaurusBaseWidget, name, designMode=designMode)

        self.title = ''
        self.showLabels = True
        self.filter = ''
        self._modelNames = []
        self.row_labels = []
        self.column_labels = []
        self._widgets_list = []
        self._last_selected = None
        self._show_frames = True
        self._show_row_frame = True
        self._show_column_frame = True
        self._show_others = False
        self._show_attr_labels = True
        self._show_attr_units = True
        self.hideLabels = False

        self.defineStyle()
        self.modelsQueue = Queue()
        self.__modelsThread = None
        if not designMode:
            self.modelsThread

    @property
    def modelsThread(self):
        modelsThread = self.__modelsThread
        if modelsThread is None:
            modelsThread = SingletonWorker(parent=self, name='TaurusGrid',
                                           queue=self.modelsQueue,
                                           method=modelSetter, cursor=True)
            self.__modelsThread = modelsThread
        return modelsThread

    def save(self, filename):
        import pickle
        d = {
            'model': self.filter,
            'row_labels': self.row_labels, 'column_labels': self.column_labels,
            'frames': self._show_row_frame or self._show_column_frame,
            'labels': self._show_attr_labels,
            'units': self._show_attr_units, 'others': self._show_others
        }
        f = open(filename, 'wb')
        pickle.dump(d, f)
        f.close()

    def load(self, filename, delayed=False):
        self.trace('In TauGrid.load(%s,%s)' % (filename, delayed))
        if not isinstance(filename, dict):
            manual = False
            import pickle
            f = open(filename, 'rb')
            d = pickle.load(f)
            f.close()
        else:
            manual = True
            d = filename
        self.setRowLabels(d['row_labels'])
        self.setColumnLabels(d['column_labels'])
        self.showAttributeLabels(d.get('labels', True))
        self.showAttributeUnits(d.get('units', True))
        self.showOthers(d.get('others', True))
        self.showRowFrame(d.get('frames', True))
        if manual:
            self.showColumnFrame(d.get('frames', True))
        self.setModel(d['model'], delayed=d.get('delayed', delayed))
        return self._modelNames

    def defineStyle(self):
        """ Defines the initial style for the widget """
        # ----------------------------------------------------------------------
        # Write your own code here to set the initial style of your widget
        #
        self.setLayout(QtGui.QGridLayout())
        # self.layout().setContentsMargins(0,0,0,0)
        self.updateStyle()

    def sizeHint(self):
        return QtGui.QFrame.sizeHint(self)

    def minimumSizeHint(self):
        return QtGui.QFrame.minimumSizeHint(self)

    # -~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-
    # TaurusBaseWidget over writing
    # -~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-

    def getModelClass(self):
        # ----------------------------------------------------------------------
        # [MANDATORY]
        # Replace your own code here
        # ex.: return taurus.core.taurusattribute.Attribute
        raise RuntimeError("Forgot to overwrite %s.getModelClass" % str(self))
        return list

    def attach(self):
        """Attaches the widget to the model"""

        if self.isAttached():
            return True

        # ----------------------------------------------------------------------
        # Write your own code here before attaching widget to attribute connect
        # the proper signal so that the first event is correctly received by the
        # widget

        ret = TaurusBaseWidget.attach(self)

        # by default enable/disable widget according to attach state
        self.setEnabled(ret)
        return ret

    def detach(self):
        """Detaches the widget from the model"""

        TaurusBaseWidget.detach(self)

        # ----------------------------------------------------------------------
        # Write your own code here after detaching the widget from the model

        # by default disable widget when dettached
        self.setEnabled(False)

    # -------------------------------------------------------------------------
    # [MANDATORY]
    # Uncomment the following method if your superclass does not provide with a
    # isReadOnly() method or if you need to change its behavior

    # def isReadOnly(self):
    #    return True

    def updateStyle(self):
        # ----------------------------------------------------------------------
        # Write your own code here to update your widget style
        self.trace('@' * 80)
        self.trace(
            'In TaurusGrid.updateStyle() ....... It seems never called!!!!')
        self.trace('@' * 80)

        # It was showing an annoying "True" in the widget
        # value = self.getShowText() or ''
        # if self._setText: self._setText(value) ##It must be included

        # update tooltip
        self.setToolTip(self.getFormatedToolTip())  # It must be included

        # send a repaint in the end
        if hasattr(self, 'title_widget'):
            if self.title:
                self.title_widget.show()
            else:
                self.title_widget.hide()

        self.update()

    # -------------------------------------------------------------------------
    # Write your own code here for your own widget properties

    def setModel(self, model, devsInRows=False, delayed=False, append=False,
                 load=True):
        '''The model can be initialized as a list of devices or hosts or dictionary or ...'''
        # self.setModelCheck(model) ##It must be included
        # differenciate if the model is a RegExp
        if isinstance(model, dict):
            self.load(model)
        else:
            model = isinstance(model, string_types) and [model] or list(model)
            self.trace('#' * 80)
            self.trace('In TaurusGrid.setModel(%s)' % str(model)[:100])

            self.delayed = delayed
            self.filter = model
            if any('*' in m for m in model):
                model = get_all_models(model)
                self.debug(
                    'model was a RegExp, done the query and converted to an attr list')

            if not self._modelNames == []:  # clean to start from scratch
                for widget in self._widgets_list:
                    del widget

            # here we always have the reals model list, even if it comes from a
            # regexp
            if append:
                self._modelNames = self._modelNames + model
            else:
                self._modelNames = model

            self.debug(('In TaurusGrid.setModel(...): modelNames are %s' %
                        (self._modelNames))[:100] + '...')

            if load:
                self.trace(
                    'In TaurusGrid.setModel(%s,load=True): modelNames are %d' % (
                        str(model)[:100] + '...',
                        len(self._modelNames)))  # ,self._modelNames))
                if devsInRows:
                    self.setRowLabels(
                        ','.join(
                            set(d.rsplit('/', 1)[0] for d in self._modelNames)))
                self.create_widgets_table(self._modelNames)
                self.modelsQueue.put(
                    (MethodModel(self.showRowFrame), self._show_row_frame))
                self.modelsQueue.put(
                    (
                    MethodModel(self.showColumnFrame), self._show_column_frame))
                self.modelsQueue.put(
                    (MethodModel(self.showOthers), self._show_others))
                self.modelsQueue.put(
                    (MethodModel(self.showAttributeLabels),
                     self._show_attr_labels))
                self.modelsQueue.put(
                    (MethodModel(self.showAttributeUnits),
                     self._show_attr_units))
                self.updateStyle()

                if not self.delayed:
                    self.trace('In setModel(): not delayed loading of models')
                    if not self.modelsThread.isRunning():
                        # print 'In setModel(): Starting Thread! (%d objs in
                        # queue)'%(self.modelsThread.queue.qsize())
                        self.trace('<' * 80)
                        # self.modelsThread.IdlePriority)
                        self.modelsThread.start()
                    else:
                        # print 'In setModel(): Thread already started! (%d
                        # objs in queue)'%(self.modelsThread.queue.qsize())
                        self.modelsThread.next()
                else:
                    self.trace('In setModel(): models loading delayed!')
                    pass

            self.trace('Out of TaurusGrid.setModel(%s)' % str(model)[:100])
            self.updateStyle()
        return

    def getModel(self):
        return self._modelNames

    def resetModel(self):
        self._modelNames = []
        self.updateFromList(self._modelNames)
        return

    def setTitle(self, title):
        self.title = str(title)
        if hasattr(self, 'title_widget'):
            if title:
                self.title_widget.setText(self.title)
                self.title_widget.show()
            else:
                self.title_widget.hide()

    def parse_labels(self, text):
        if any(text.startswith(c[0]) and text.endswith(c[1]) for c in
               [('{', '}'), ('(', ')'), ('[', ']')]):
            try:
                labels = eval(text)
                return labels
            except Exception as e:
                self.warning(
                    'ERROR! Unable to parse labels property: %s' % str(e))
                return []
        else:
            exprs = [t.strip() for t in text.split(',')]
            labels = [(':' in e and (e.split(':', 1)[0].strip(), e.split(
                ':', 1)[-1].strip()) or (e, e)) for e in exprs]
            return labels

    def setRowLabels(self, rows):
        '''The model can be initialized as a list of devices or hosts or ...'''
        # self.setModelCheck(model) ##It must be included
        self.row_labels = self.parse_labels(str(rows))
        try:
            self.rows = [r[0] for r in self.row_labels]
            for i in range(len(self.rows)):
                section = self.rows[i]
                self.table.setVerticalHeaderItem(
                    i, QtGui.QTableWidgetItem(section))
        except Exception as e:
            self.debug("setRowLabels(): Exception! %s" % e)
            # self.create_widgets_table(self._columnsNames)

    def getRowLabels(self):
        return ','.join(':'.join(c) for c in self.row_labels)

    def resetRowLabels(self):
        self.row_labels = []
        return

    def setColumnLabels(self, columns):
        '''The model can be initialized as a list of devices or hosts or ...'''
        # self.setModelCheck(model) ##It must be included
        self.column_labels = self.parse_labels(str(columns))
        try:
            self.columns = [c[0] for c in self.column_labels]
            for i in range(len(self.columns)):
                equipment = self.columns[i]
                self.table.setHorizontalHeaderItem(
                    i, QtGui.QTableWidgetItem(equipment))
        except Exception as e:
            self.debug("setColumnLabels(): Exception! %s" % e)
            # self.create_widgets_table(self._columnsNames)

    def getColumnLabels(self):
        return ','.join(':'.join(c) for c in self.column_labels)

    def resetColumnLabels(self):
        self.column_labels = []
        return

    # FIXME: when they are called before setModel they fails because
    # frames are not yet created, and it doesn't has memoty about this.
    def showRowFrame(self, boolean):
        self._show_row_frame = boolean
        if hasattr(self, 'rows_frame'):
            if boolean:
                self.rows_frame.show()
            else:
                self.rows_frame.hide()

    def showColumnFrame(self, boolean):
        self._show_column_frame = boolean
        if hasattr(self, 'columns_frame'):
            if boolean:
                self.columns_frame.show()
            else:
                self.columns_frame.hide()

    def showAttributeLabels(self, boolean):
        self.trace('In showAttributeLabels(%s)' % boolean)
        self._show_attr_labels = boolean
        for tv in self._widgets_list:
            try:
                if tv and tv.labelWidget:
                    if boolean:
                        tv.labelWidget().show()
                    else:
                        tv.labelWidget().hide()
            except:
                pass
        return self._show_attr_labels

    def showAttributeUnits(self, boolean):
        self.trace('In showAttributeUnits(%s)' % boolean)
        self._show_attr_units = boolean
        for tv in self._widgets_list:
            try:
                if tv and tv.unitsWidget:
                    if boolean:
                        tv.unitsWidget().show()
                    else:
                        tv.unitsWidget().hide()
            except:
                pass
        return self._show_attr_units

    # -~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-
    # QT properties
    # -~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-

    model = QtCore.pyqtProperty("QStringList", getModel,
                                setModel,
                                resetModel)

    rowlabels = QtCore.pyqtProperty("QString", getRowLabels,
                                    setRowLabels,
                                    resetRowLabels)

    columnlabels = QtCore.pyqtProperty("QString", getColumnLabels,
                                       setColumnLabels,
                                       resetColumnLabels)

    useParentModel = QtCore.pyqtProperty("bool",
                                         TaurusBaseWidget.getUseParentModel,
                                         TaurusBaseWidget.setUseParentModel,
                                         TaurusBaseWidget.resetUseParentModel)

    # -------------------------------------------------------------------------
    # Write your own code here for your own widget properties

    def create_widgets_dict(self, models):
        from collections import defaultdict
        # recursive dictionary with 2 levels
        values = defaultdict(lambda: defaultdict(list))
        # domains = list(set(m.split('/')[0].upper()))
        # families = list(set(m.split('/')[1].upper()))

        if not self.row_labels:  # Domains used by default
            self.row_labels = sorted(
                list(set(m.split('/')[0].upper() for m in models if
                         m.count('/') >= 2)))
            self.row_labels = list(zip(self.row_labels, self.row_labels))
        if not self.column_labels:  # Families used by default
            self.column_labels = sorted(
                list(set(m.split('/')[1].upper() for m in models if
                         m.count('/') >= 2)))
            self.column_labels = list(zip(self.column_labels, self.column_labels))

            # for m in models:
            # if m.count('/')<2:
            # self.warning('Wrong model cannot be added: %s'%m)
            # else:
            # domain,family = m.upper().split('/')[:2]
            # values[domain][family].append(m)

        row_not_found, col_not_found = False, False

        for m in models:
            row, column = 'Others', 'Others'
            for label, rexp in self.row_labels:
                if '*' in rexp and '.*' not in rexp:
                    rexp = rexp.replace('*', '.*')
                if re_search_low(rexp, m):
                    row = label
                    break
            for label, rexp in self.column_labels:
                if '*' in rexp and '.*' not in rexp:
                    rexp = rexp.replace('*', '.*')
                if re_search_low(rexp, m):
                    column = label
                    break
            if 'Others' == row:
                row_not_found = True
            if 'Others' == column:
                col_not_found = True
            self.debug('Model %s added to row %s , column %s' %
                       (m, row, column))
            values[row][column].append(m)
        if row_not_found:
            self.row_labels.append(('Others', '.*'))
        if col_not_found:
            self.column_labels.append(('Others', '.*'))

        return values

    def create_frame_with_gridlayout(self):
        """ Just a 'macro' to create the layouts that seem to fit better. """
        frame = TaurusGrid._TaurusGridCell()
        frame.setLayout(QtGui.QGridLayout())
        frame.layout().setContentsMargins(2, 2, 2, 2)
        frame.layout().setSpacing(0)
        frame.layout().setSpacing(0)
        return frame

    def create_widgets_table(self, models):

        # Added a title to the panel
        self.title_widget = QtGui.QLabel()
        self.layout().addWidget(self.title_widget, 0, 0)
        self.setTitle(self.title)

        dct = self.create_widgets_dict(models)
        # Assignments modified to keep the order of labels as inserted in
        # properties
        self.rows = [r[0] for r in self.row_labels]  # dct.keys()
        # list(set(reduce(operator.add,v.keys()) for v in dct.values()))
        self.columns = [c[0] for c in self.column_labels]

        values = []
        for row in self.rows:
            line = []
            for col in self.columns:
                # line.append(dct[row][col] if col in dct[row] else [])
                if col in dct[row]:
                    line.append(dct[row][col])
                else:
                    line.append([])
            values.append(line)

        # Here is where the table is created!
        self.table = self.build_table(values)

        # SET COLUMN HEADERS (self.columns)
        for i in range(len(self.columns)):
            equipment = self.columns[i]
            self.table.setHorizontalHeaderItem(
                i, QtGui.QTableWidgetItem(equipment))
            # SOMEDAY THIS WILL BE ICONS

        # SET ROW HEADERS (self.rows)
        for i in range(len(self.rows)):
            section = self.rows[i]
            self.table.setVerticalHeaderItem(
                i, QtGui.QTableWidgetItem(section))

        #        table.setAutoScroll(True)
        #        resize_mode = QtGui.QHeaderView.Stretch
        #        table.horizontalHeader().setSectionResizeMode(resize_mode)
        # table.verticalHeader().setSectionResizeMode(resize_mode)

        #        for row in range(len(self.rows)):
        #            table.setRowHeight(row,5+25*sum(len(dct[self.rows[row]][col]) for col in self.columns))
        # for col in range(len(self.columns)):
        #    table.setColumnWidth(col,300)

        use_scroll = False  # It didn't work ... it doesn't allow the table widget to resize
        if use_scroll:
            scrollable = QtGui.QScrollArea(self)
            scrollable.setWidget(self.table)
            self.layout().addWidget(scrollable, 1, 0)
        else:
            self.layout().addWidget(self.table, 1, 0)

        # ----------------------------------------------------------------------
        # SECTION CHECKBOXES
        self.checkboxes_frame = self.create_frame_with_gridlayout()

        self.rows_frame = self.create_frame_with_gridlayout()
        self.rows_frame.setFrameStyle(QtGui.QFrame.Box)
        if not self._show_row_frame:
            self.rows_frame.hide()
        self.checkboxes_frame.layout().addWidget(self.rows_frame, 0, 0)

        self.columns_frame = self.create_frame_with_gridlayout()
        self.columns_frame.setFrameStyle(QtGui.QFrame.Box)
        if not self._show_column_frame:
            self.columns_frame.hide()
        self.checkboxes_frame.layout().addWidget(self.columns_frame, 0, 1)

        layout_row = 0
        layout_col = 0
        # For all rows, create rows of three checkboxes in order to
        # show or hide the corresponding rows in the table
        for i in range(len(self.rows)):
            section = self.rows[i]
            checkbox = QtGui.QCheckBox(section)

            # -------------------------------------------------------
            # Work around for https://bugs.kde.org/show_bug.cgi?id=345023
            # TODO: make better solution for this
            checkbox._id = section  # <-- ugly monkey-patch!
            # -------------------------------------------------------

            if section == 'Others':
                checkbox.setChecked(False)
                if not self._show_others:
                    checkbox.hide()
            else:
                checkbox.setChecked(True)
            self.rows_frame.layout().addWidget(checkbox, layout_row, layout_col)
            layout_col += 1
            if layout_col == 3:
                layout_col = 0
                layout_row += 1
            checkbox.toggled.connect(self.show_hide_rows)
        self.show_hide_rows()

        layout_row = 0
        layout_col = 0
        # For all self.columns, create rows of three checkboxes in order to
        # show or hide the corresponding columns in the table
        for i in range(len(self.columns)):
            column = self.columns[i]
            checkbox = QtGui.QCheckBox(column)

            # -------------------------------------------------------
            # Work around for https://bugs.kde.org/show_bug.cgi?id=345023
            # TODO: make better solution for this
            checkbox._id = column  # <-- ugly monkey-patch!
            # -------------------------------------------------------

            if column == 'Others':
                checkbox.setChecked(False)
                if not self._show_others:
                    checkbox.hide()
            else:
                checkbox.setChecked(True)
            self.columns_frame.layout().addWidget(checkbox, layout_row,
                                                  layout_col)
            layout_col += 1
            if layout_col == 3:
                layout_col = 0
                layout_row += 1
            checkbox.toggled.connect(self.show_hide_columns)
        self.show_hide_columns()

        self.layout().addWidget(self.checkboxes_frame, 2, 0)
        # self.resize(800,600)

    def show_hide_rows(self):
        """
        This needs refactoring to be together with the show_hide_columns method
        """
        for checkbox in self.rows_frame.children():
            if isinstance(checkbox, QtGui.QCheckBox):
                # -------------------------------------------------------
                # Work around for https://bugs.kde.org/show_bug.cgi?id=345023
                # TODO: make better solution for this
                # checkbox.text()  <-- fails due to added "&
                table_row = self.rows.index(checkbox._id)
                # -------------------------------------------------------
                if checkbox.isChecked():
                    self.table.showRow(table_row)
                else:
                    self.table.hideRow(table_row)

    def show_hide_columns(self):
        """
        This needs refactoring to be together with the show_hide_rows method
        """
        for checkbox in self.columns_frame.children():
            if isinstance(checkbox, QtGui.QCheckBox):
                # -------------------------------------------------------
                # Work around for https://bugs.kde.org/show_bug.cgi?id=345023
                # TODO: make better solution for this
                # checkbox.text()  <-- fails due to added "&
                table_col = self.columns.index(checkbox._id)
                # -------------------------------------------------------
                if checkbox.isChecked():
                    self.table.showColumn(table_col)
                else:
                    self.table.hideColumn(table_col)

    def showOthers(self, boolean):
        self._show_others = boolean
        if hasattr(self, 'rows_frame'):
            for checkbox in self.rows_frame.children():
                if (isinstance(checkbox, QtGui.QCheckBox)
                # -------------------------------------------------------
                # Work around for https://bugs.kde.org/show_bug.cgi?id=345023
                # TODO: make better solution for this
                # checkbox.text()  <-- fails due to added "&
                        and checkbox._id == 'Others'):
                # -------------------------------------------------------
                    if self._show_others:
                        checkbox.show()
                    else:
                        checkbox.hide()
        if hasattr(self, 'columns_frame'):
            for checkbox in self.columns_frame.children():
                if (isinstance(checkbox, QtGui.QCheckBox)
                # -------------------------------------------------------
                # Work around for https://bugs.kde.org/show_bug.cgi?id=345023
                # TODO: make better solution for this
                # checkbox.text()  <-- fails due to added "&
                        and checkbox._id == 'Others'):
                # -------------------------------------------------------
                    if self._show_others:
                        checkbox.show()
                    else:
                        checkbox.hide()

    def build_table(self, values):
        """
        This is a builder. For all the elements in widgets matrix,
        just set the corresponding cells of the QTableWidget.
        """
        self.trace('In TaurusGrid.build_table(%s)' % values)
        widgets_matrix = self.build_widgets(values, self.showLabels)
        rows = len(widgets_matrix)
        cols = rows and len(widgets_matrix[0]) or 0

        table = QtGui.QTableWidget()
        table.setItemDelegate(Delegate(table))
        # This example replaces the blue background of selected cells in tables
        palette = Qt.QPalette()
        palette.setBrush(palette.Active, palette.Highlight,
                         Qt.QBrush(Qt.Qt.white))
        table.setPalette(palette)

        table.setRowCount(rows)
        table.setColumnCount(cols)
        for row in range(len(widgets_matrix)):
            for col in range(len(widgets_matrix[row])):
                table.setCellWidget(row, col, widgets_matrix[row][col])

        # table.resizeColumnsToContents()
        # table.resizeRowsToContents()
        hh = table.horizontalHeader()
        if hh.length() > 0:
            try:
                hh.setSectionResizeMode(hh.Stretch)
            except AttributeError:  # PyQt4
                hh.setResizeMode(hh.Stretch)
        # table.verticalHeader().setSectionResizeMode(QtGui.QHeaderView.Stretch)
        # table.horizontalHeader().setSectionResizeMode(QtGui.QHeaderView.ResizeToContents)
        vh = table.verticalHeader()
        if vh.length() > 0:
            try:
                vh.setSectionResizeMode(vh.ResizeToContents)
            except AttributeError:  # PyQt4
                hh.setResizeMode(vh.ResizeToContents)

        return table

    def build_widgets(self, values, show_labels=False, width=240, height=20,
                      value_width=120):
        widgets_matrix = []
        for row in values:
            widgets_row = []
            for cell in row:
                cell_frame = self.create_frame_with_gridlayout()
                cell_frame.itemClicked.connect(self.onItemClicked)
                count = 0
                for synoptic in sorted(cell):
                    self.debug("processing synoptic %s" % synoptic)
                    name = model = synoptic

                    self.debug('Creating TaurusValue with model =  %s' % model)
                    synoptic_value = TaurusValue(cell_frame)
                    self.modelsQueue.put((synoptic_value, model))

                    if self.hideLabels:
                        synoptic_value.setLabelWidgetClass(None)
                    else:
                        # DO NOT DELETE THIS LINE!!!
                        synoptic_value.setLabelConfig('label')
                    cell_frame.layout().addWidget(synoptic_value, count, 0)
                    self._widgets_list.append(synoptic_value)
                    count += 1

                widgets_row.append(cell_frame)
            widgets_matrix.append(widgets_row)
        return widgets_matrix

    def onItemClicked(self, item_name):
        self.trace('In TaurusGrid.itemClicked(%s)' % item_name)
        self.setItemSelected(item_name)
        self.itemClicked.emit(str(item_name))

    def setItemSelected(self, item_name='', selected=True):
        """ it adds a blue frame around a clicked item. """
        if isinstance(item_name, TaurusValue):
            self.trace('In TaurusGrid.setItemSelected(%s,%s)' %
                       (str(item_name.getModel()), selected))
            item = item_name
        else:
            self.trace('In TaurusGrid.setItemSelected(%s,%s)' %
                       (str(item_name), selected))
            if item_name:
                item = self.getItemByModel(item_name)
            else:
                item = self._last_selected
        if item:
            if selected:
                item._labelWidget.setStyleSheet(
                    'border-style: solid ; border-width: 1px; border-color: blue; color: blue; border-radius:4px;')
                if self._last_selected and self._last_selected != item:
                    self.setItemSelected(self._last_selected, False)
                self._last_selected = item
            else:
                item._labelWidget.setStyleSheet(
                    'border-style: solid; border-width: 1px; border-color: transparent; color: black;  border-radius:4px;')
                self._last_selected = None
        else:
            return None

    def getItemByModel(self, model, index=0):
        # a particular model can be many times, index means which one of them
        model = str(model).lower()
        for widget in self._widgets_list:
            if str(widget.getModel()).lower() == model:
                if index <= 0:
                    return widget
                else:
                    index -= 1

    @classmethod
    def getQtDesignerPluginInfo(cls):
        ret = TaurusBaseWidget.getQtDesignerPluginInfo()
        ret['module'] = 'taurus.qt.qtgui.table'
        ret['group'] = 'Taurus Views'
        ret['icon'] = "designer:grid.png"
        return ret


class Delegate(QtGui.QItemDelegate):
    def __init__(self, parent=None):
        QtGui.QItemDelegate.__init__(self, parent)

    def sizeHint(self, option, index):
        table = self.parent()
        widget = table.cellWidget(index.row(), index.column())
        size = widget.sizeHint()
        return size


def sysargs_to_dict(defaults=[]):
    import sys
    i, result = 0, {}
    for a in sys.argv[1:]:
        if '=' in a:
            bar = a.split('=')
            if bar[1] in ['True', 'False']:  # convert string to boolean
                bar[1] = eval(bar[1])
            result[bar[0].replace('--', '')] = bar[1]
        else:
            result[defaults[i]] = a
            i += 1
    return result


if __name__ == '__main__':
    import sys
    from taurus.qt.qtgui.application import TaurusApplication

    if len(sys.argv) < 2:
        print("The format of the call is something like:")
        print('\t/usr/bin/python taurusgrid.py grid.pickle.file')
        print('\t/usr/bin/python taurusgrid.py "model=lt.*/VC.*/.*/((C*)|(P*)|(I*))" cols=IP,CCG,PNV rows=LT01,LT02 others=False rowframe=True colframe=False')
        exit()

    app = TaurusApplication(sys.argv[0:1], cmd_line_parser=None)
    gui = TaurusGrid()

    try:  # first try if argument is a file to be opened
        filename = sys.argv[1]
        open(filename, 'r')
        gui.load(filename)
    except:
        args = sysargs_to_dict(
            ['model', 'rows', 'cols', 'others', 'rowframe', 'colframe'])
        print("args = %s" % args)
        if args.get('rows'):
            gui.setRowLabels(args['rows'])
        if args.get('cols'):
            gui.setColumnLabels(args['cols'])
        if args.get('model'):
            gui.setModel(args['model'])
        gui.showRowFrame('rowframe' in args and args['rowframe'] and True)
        gui.showColumnFrame('colframe' in args and args['colframe'] and True)
        gui.showOthers('others' in args and args['others'] or True)

    print("current TaurusGrid model= %s" % (gui.getModel()))
    gui.show()
    sys.exit(app.exec_())



#!/usr/bin/env python

#############################################################################
##
# This file is part of Taurus
##
# http://taurus-scada.org
##
# Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
# Taurus is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
##
# Taurus is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
##
# You should have received a copy of the GNU Lesser General Public License
# along with Taurus.  If not, see <http://www.gnu.org/licenses/>.
##
#############################################################################
"""
taurusgraphic.py:
"""
from __future__ import print_function

# TODO: Tango-centric

from future import standard_library
standard_library.install_aliases()
from builtins import str
from builtins import range
from builtins import object

import re
import os
import subprocess
import traceback
try:
    from collections.abc import Sequence
except ImportError:  # bck-compat py 2.7
    from collections import Sequence

from future.utils import string_types
from queue import Queue

from taurus import Manager
from taurus.core import AttrQuality, DataType
from taurus.core.util.containers import CaselessDefaultDict
from taurus.core.util.log import Logger, deprecation_decorator
from taurus.core.taurusdevice import TaurusDevice
from taurus.core.taurusattribute import TaurusAttribute
from taurus.core.util.enumeration import Enumeration
from taurus.external.qt import Qt, compat
from taurus.qt.qtgui.base import TaurusBaseComponent
from taurus.qt.qtgui.util import (QT_ATTRIBUTE_QUALITY_PALETTE, QT_DEVICE_STATE_PALETTE,
                                  ExternalAppAction, TaurusWidgetFactory)


__all__ = ['SynopticSelectionStyle',
           'parseTangoUri',
           'QEmitter',  # TODO: QEmitter should probably be removed (kept priv)
           'TaurusGraphicsUpdateThread',
           'TaurusGraphicsScene',
           'QGraphicsTextBoxing',
           'QSpline',
           'TaurusGraphicsItem',
           'TaurusGraphicsAttributeItem',
           'TaurusGraphicsStateItem',
           'TaurusEllipseStateItem',
           'TaurusRectStateItem',
           'TaurusSplineStateItem',
           'TaurusRoundRectItem',
           'TaurusRoundRectStateItem',
           'TaurusGroupItem',
           'TaurusGroupStateItem',
           'TaurusPolygonStateItem',
           'TaurusLineStateItem',
           'TaurusTextStateItem',
           'TaurusTextAttributeItem',
           'TYPE_TO_GRAPHICS',
           'TaurusBaseGraphicsFactory',
           ]

__docformat__ = 'restructuredtext'


SynopticSelectionStyle = Enumeration("SynopticSelectionStyle", [
    # A blue ellipse is displayed around the selected objects
    "ELLIPSE",
    # The own outline of selected object is displayed in blue and bolder
    "OUTLINE",
])


def parseTangoUri(name):
    # TODO: Tango-centric
    from taurus.core import tango
    from taurus.core.tango.tangovalidator import (TangoDeviceNameValidator,
                                                  TangoAttributeNameValidator)

    validator = {tango.TangoDevice: TangoDeviceNameValidator,
                 tango.TangoAttribute: TangoAttributeNameValidator}
    try:
        val = validator[tango.TangoFactory().findObjectClass(name)]()
        params = val.getUriGroups(name)
        return (params if '_devslashname' in params else None)
    except:
        return None


class QEmitter(Qt.QObject):
    updateView = Qt.pyqtSignal(compat.PY_OBJECT)


class TaurusGraphicsUpdateThread(Qt.QThread):

    def __init__(self, parent=None, period=3):
        """Parent most not be None and must be a TaurusGraphicsScene!"""
        if not isinstance(parent, TaurusGraphicsScene):
            raise RuntimeError("Illegal parent for TaurusGraphicsUpdateThread")
        Qt.QThread.__init__(self, parent)
        self.period = period
        self.log = Logger('TaurusGraphicsUpdateThread')

    def _updateView(self, v):
        # The first one is the prefered one because it improves performance
        # since updates don't come very often in comparison to with the refresh
        # rate of the monitor (~70Hz)
        if v.viewportUpdateMode() == Qt.QGraphicsView.NoViewportUpdate:
            # We call the update to the viewport instead of the view
            # itself because apparently there is a bug in QT 4.3 that
            # prevents a proper update when the view is inside a QTab
            v.viewport().update()
        else:
            # @todo This is probably a bug (item_rects is not defined). But it is defined in .run(), see "todo" below...
            v.updateScene(item_rects)
            # v.invalidateScene(item.boundingRect())
        return

    def run(self):
        self.log.debug("run... - TaurusGraphicsUpdateThread")
        emitter = QEmitter()
        emitter.moveToThread(Qt.QApplication.instance().thread())
        emitter.setParent(Qt.QApplication.instance())
        emitter.updateView.connect(self._updateView)

        p = self.parent()
        while True:
            item = p.getQueue().get(True)
            if type(item) in (str,):
                if item == "exit":
                    break
                else:
                    continue
            if not isinstance(item, Sequence):
                item = (item,)
            # @todo: Unless the call to boundingRect() has a side effect, this line is useless..  probably related to todo in _updateView()
            item_rects = [i.boundingRect() for i in item]

            for v in p.views():
                # p.debug("emit('updateView')")
                # emitter.updateView.emit(v)
                emitter.updateView.emit(v)
            # This sleep is needed to reduce CPU usage of the application!
            self.sleep(self.period)
            # End of while
        # End of Thread


class TaurusGraphicsScene(Qt.QGraphicsScene):
    '''
    This class encapsulates TaurusJDrawSynopticsView and TaurusGraphicsScene signals/slots

    External events::

     Slot selectGraphicItem(const QString &) displays a selection
     mark around the TaurusGraphicsItem that matches the argument passed.

    Mouse Left-button events::

     Signal graphicItemSelected(QString) is triggered, passing the
     selected TaurusGraphicsItem.name() as argument.

    Mouse Right-button events::

     TaurusGraphicsItem.setContextMenu([(ActionName,ActionMethod(device_name))]
     allows to configure custom context menus for graphic items using a list
     of tuples. Empty tuples will insert separators in the menu.
    '''
    ANY_ATTRIBUTE_SELECTS_DEVICE = True
    TRACE_ALL = False

    refreshTree2 = Qt.pyqtSignal()
    graphicItemSelected = Qt.pyqtSignal('QString')
    graphicSceneClicked = Qt.pyqtSignal('QPoint')

    def __init__(self, parent=None, strt=True):
        name = self.__class__.__name__
        # self.call__init__(Logger, name, parent) #Inheriting from Logger
        # caused exceptions in CONNECT
        Qt.QGraphicsScene.__init__(self, parent)
        self.updateQueue = None
        self.updateThread = None
        self._itemnames = CaselessDefaultDict(lambda k: set())
        self._selection = []
        self._selectedItems = []
        self._selectionStyle = SynopticSelectionStyle.OUTLINE
        self.threads = []
        self.pids = []
        self.panels = []
        self.panel_launcher = None

        try:
            self.logger = Logger(name)
            # self.logger.setLogLevel(self.logger.Info)
            if not self.TRACE_ALL:
                self.debug = lambda l: self.logger.debug(l)
                self.info = lambda l: self.logger.info(l)
                self.warning = lambda l: self.logger.warning(l)
            else:
                self.debug = self.info = self.warning = self.error = lambda l: self.logger.warning(
                    l)
        except:
            print('Unable to initialize TaurusGraphicsSceneLogger: %s' % traceback.format_exc())

        try:
            if parent and parent.panelClass() is not None:
                defaultClass = parent.panelClass()
                if defaultClass and isinstance(defaultClass, string_types):
                    self.panel_launcher = self.getClass(defaultClass)
                    if self.panel_launcher is None:
                        self.panel_launcher = ExternalAppAction(
                            defaultClass.split())
                else:
                    self.panel_launcher = defaultClass
            else:
                from taurus.qt.qtgui.graphic.jdraw import TaurusJDrawSynopticsView
                self.panel_launcher = TaurusJDrawSynopticsView.defaultPanelClass()
        except:
            self.warning(traceback.format_exc())
            self.panel_launcher = None

        self.setSelectionMark()
        if strt:
            self.start()

    def __del__(self):
        self.closeAllPanels()
        Qt.QGraphicsScene.__del__(self)

    def showNewPanel(self, args=None, standAlone=False):
        try:
            if isinstance(args, TaurusGraphicsItem):
                objName = args._name
                clName = args.getExtensions().get('className') or self.panel_launcher
                # classParams extension overrides Model; if there's no
                # extension then object name is used
                clParam = args.getExtensions().get('classParams') or objName
                standAlone = args.standAlone
            else:
                clName, clParam, objName = self.panel_launcher, args, args
            if not clName or clName == 'noPanel':
                return
            self.debug('TaurusGraphicsScene.showNewPanel(%s,%s,%s)' %
                       (clName, clParam, objName))
            if isinstance(clName, ExternalAppAction):
                clName.actionTriggered(clParam if isinstance(
                    clParam, (list, tuple)) else [clParam])
            else:
                if isinstance(clName, string_types):
                    klass = self.getClass(clName)
                    if klass is None:
                        self.warning("%s Class not found!" % clName)
                        klass = self.getClass("TaurusDevicePanel")
                else:
                    klass, clName = clName, getattr(
                        clName, '__name__', str(clName))
                widget = klass()  # self.parent())
                # if isinstance(widget,taurus.qt.qtgui.panel.TaurusDevicePanel):
                # widget.setSpectraAtkMode(True) #Method renamed or deprecated
                try:
                    widget.setClasses(clParam)
                except:
                    pass
                try:
                    widget.setModel(clParam)
                except:
                    pass
                try:
                    widget.setTable(clParam)
                except:
                    pass

                # if isinstance(widget,Qt.QWidget):
                    # if not standAlone:
                        #obj = newDialog(self.parent())
                    # else:
                        #obj = newDialog()
                    # obj.initComponents(widget,objName,clName)
                    # obj.setModal(False)
                    # obj.setVisible(True)

                widget.setWindowTitle('%s - %s' % (clName, objName))
                self.panels.append(widget)
                widget.show()  # exec_()
                return widget
        except:
            self.warning(traceback.format_exc())

    def closeAllPanels(self):
        """ This method replaces killProcess, using taurus.qt.qtgui.util.ExternalAppAction instead! """
        try:
            self.debug('In closeAllPanels(%s,%s)' %
                       (self.panel_launcher, self.panels))
            if isinstance(self.panel_launcher, ExternalAppAction):
                self.panel_launcher.kill()
            for p in self.panels:
                try:
                    if hasattr(p, 'setModel'):
                        p.setModel(None)
                    p.close()
                except:
                    pass
            while self.panels:
                self.panels.pop(0)
        except:
            self.warning(traceback.format_exc())

    def addItem(self, item):
        # self.debug('addItem(%s)'%item)
        def expand(i):
            name = str(getattr(i, '_name', '')).lower()
            if name:
                self._itemnames[name].add(i)
                #self.debug('addItem(%s): %s'%(name,i))
            if isinstance(i, Qt.QGraphicsItemGroup):
                for j in i.childItems():
                    expand(j)
        expand(item)
        Qt.QGraphicsScene.addItem(self, item)

    def addWidget(self, item, flags=None):
        self.debug('addWidget(%s)' % item)
        name = str(getattr(item, '_name', '')).lower()
        if name:
            self._itemnames[name].add(item)
        if flags is None:
            Qt.QGraphicsScene.addWidget(self, item)
        else:
            Qt.QGraphicsScene.addWidget(self, item, flags)

    def getItemByName(self, item_name, strict=None):
        """
        Returns a list with all items matching a given name.

        :param strict: (bool or None) controls whether full_name (strict=True) or only device name (False) must match

        :return: (list) items
        """

        # TODO: Tango-centric
        from taurus.core.tango.tangovalidator import (
            TangoDeviceNameValidator, TangoAttributeNameValidator)

        strict = (
            not self.ANY_ATTRIBUTE_SELECTS_DEVICE) if strict is None else strict
        alnum = r'(?:[a-zA-Z0-9-_\*]|(?:\.\*))(?:[a-zA-Z0-9-_\*]|(?:\.\*))*'
        target = str(item_name).strip().split()[0].lower().replace(
            '/state', '')  # If it has spaces only the first word is used
        # Device names should match also its attributes or only state?
        if not strict and TangoAttributeNameValidator().getUriGroups(target):
            target = target.rsplit('/', 1)[0]
        if TangoDeviceNameValidator().getUriGroups(target):
            if strict:
                target += '(/state)?'
            else:
                target += '(/' + alnum + ')?'
        if not target.endswith('$'):
            target += '$'
        result = []
        for k in list(self._itemnames.keys()):
            if re.match(target.lower(), k.lower()):
                #self.debug('getItemByName(%s): _itemnames[%s]: %s'%(target,k,self._itemnames[k]))
                result.extend(self._itemnames[k])
        return result

    def getItemByPosition(self, x, y):
        """ This method will try first with named objects; if failed then with itemAt """
        pos = Qt.QPointF(x, y)
        itemsAtPos = []
        for z, o in sorted((i.zValue(), i) for v in self._itemnames.values() for i in v if i.contains(pos) or i.isUnderMouse()):
            if not hasattr(o, 'getExtensions'):
                self.debug(
                    'getItemByPosition(%d,%d): adding Qt primitive %s' % (x, y, o))
                itemsAtPos.append(o)
            elif not o.getExtensions().get('noSelect'):
                self.debug(
                    'getItemByPosition(%d,%d): adding GraphicsItem %s' % (x, y, o))
                itemsAtPos.append(o)
            else:
                self.debug(
                    'getItemByPosition(%d,%d): object ignored, %s' % (x, y, o))
        if itemsAtPos:
            obj = itemsAtPos[-1]
            return self.getTaurusParentItem(obj) or obj
        else:
            # return self.itemAt(x,y)
            self.debug('getItemByPosition(%d,%d): no items found!' % (x, y))
            return None

    def getItemClicked(self, mouseEvent):
        pos = mouseEvent.scenePos()
        x, y = pos.x(), pos.y()
        self.graphicSceneClicked.emit(Qt.QPoint(x, y))
        obj = self.getItemByPosition(x, y)
        #self.debug('mouse clicked on %s(%s) at (%s,%s)'%(type(obj).__name__,getattr(obj,'_name',''),x,y))
        return obj

    def mousePressEvent(self, mouseEvent):
        #self.debug('In TaurusGraphicsScene.mousePressEvent(%s,%s))'%(str(type(mouseEvent)),str(mouseEvent.button())))
        try:
            obj = self.getItemClicked(mouseEvent)
            obj_name = getattr(obj, '_name', '')
            if not obj_name and isinstance(obj, QGraphicsTextBoxing):
                obj_name = obj.toPlainText()
            if (mouseEvent.button() == Qt.Qt.LeftButton):
                # A null obj_name should deselect all, we don't send obj
                # because we want all similar to be matched
                if self.selectGraphicItem(obj_name):
                    self.debug(
                        ' => graphicItemSelected(QString)(%s)' % obj_name)
                    self.graphicItemSelected.emit(obj_name)
                else:
                    # It should send None but the signature do not allow it
                    self.graphicItemSelected.emit("")

            def addMenuAction(menu, k, action, last_was_separator=False):
                try:
                    if k:
                        configDialogAction = menu.addAction(k)
                        if action:
                            configDialogAction.triggered.connect(
                                lambda dev=obj_name, act=action: act(dev))
                        else:
                            configDialogAction.setEnabled(False)
                        last_was_separator = False
                    elif not last_was_separator:
                        menu.addSeparator()
                        last_was_separator = True
                except Exception as e:
                    self.warning('Unable to add Menu Action: %s:%s' % (k, e))
                return last_was_separator
            if (mouseEvent.button() == Qt.Qt.RightButton):
                ''' This function is called when right clicking on TaurusDevTree area. A pop up menu will be shown with the available options. '''
                self.debug('RightButton Mouse Event on %s' % (obj_name))
                if isinstance(obj, TaurusGraphicsItem) and (obj_name or obj.contextMenu() or obj.getExtensions()):
                    menu = Qt.QMenu(None)  # self.parent)
                    last_was_separator = False
                    extensions = obj.getExtensions()
                    if obj_name and (not extensions or not extensions.get('className')):
                        # menu.addAction(obj_name)
                        addMenuAction(menu, 'Show %s panel' % obj_name,
                                      lambda x=obj_name: self.showNewPanel(x))
                    if obj.contextMenu():
                        if obj_name:
                            menu.addSeparator()
                            last_was_separator = True
                        for t in obj.contextMenu():  # It must be a list of tuples (ActionName,ActionMethod)
                            last_was_separator = addMenuAction(
                                menu, t[0], t[1], last_was_separator)
                    if extensions:
                        if not menu.isEmpty():
                            menu.addSeparator()
                        className = extensions.get('className')
                        if className and className != 'noPanel':
                            self.debug('launching className extension object')
                            addMenuAction(menu, 'Show %s' % className,
                                          lambda d, x=obj: self.showNewPanel(x))
                        if extensions.get('shellCommand'):
                            addMenuAction(menu, 'Execute', lambda d,
                                          x=obj: self.getShellCommand(x))
                    if not menu.isEmpty():
                        menu.exec_(
                            Qt.QPoint(mouseEvent.screenPos().x(), mouseEvent.screenPos().y()))
                    del menu
        except Exception:
            self.warning(traceback.format_exc())

    def mouseDoubleClickEvent(self, event):
        try:
            obj = self.getItemClicked(event)
            obj_name = getattr(obj, '_name', '')
            try:
                class_name = obj.getExtensions().get('className')
            except:
                class_name = 'noPanel'
            self.debug('Clicked (%s,%s,%s)' % (obj, obj_name, class_name))
            if obj_name and class_name != 'noPanel':
                self.showNewPanel(obj)
        except:
            self.warning(traceback.format_exc())

    def setSelectionStyle(self, selectionStyle):
        # TODO We should test that selectionStyle is part of
        # SynopticSelectionStyle but there is nothing about it
        self._selectionStyle = selectionStyle

    def selectGraphicItem(self, item_name):
        """
        A blue circle is drawn around the matching item name.
        If the item_name is empty, or it is a reserved keyword, or it has the "noSelect" extension, then the blue circle is removed from the synoptic.
        """
        #self.debug('In TaurusGraphicsScene.selectGraphicItem(%s))'%item_name)
        selected = [str(getattr(item, '_name', item))
                    for item in self._selectedItems if item]
        if selected:
            iname = str(getattr(item_name, '_name', item_name))
            #self.debug('In TauGraphicsScene.selectGraphicItem(%s): already selected: %s'%(iname,selected))
            if not iname.strip():
                self.clearSelection()
                return False
            elif any(iname not in i for i in selected):
                self.clearSelection()
            else:
                self.debug(
                    'In TauGraphicsScene.selectGraphicItem(%s): already selected!' % item_name)
                return True
        if any(isinstance(item_name, t) for t in (TaurusGraphicsItem, Qt.QGraphicsItem)):
            if not getattr(item_name, '_name', ''):
                self.debug(
                    'In TauGraphicsScene.selectGraphicItem(%s): item name not found.' % item_name)
                return False
            items = [item_name]
        else:
            from taurus.qt.qtgui.graphic import jdraw_parser
            if not item_name or (str(item_name).startswith('JD') and str(item_name) in jdraw_parser.reserved):
                self.debug(
                    'In TauGraphicsScene.selectGraphicItem(%s): item name not found or name is a reserved keyword.' % item_name)
                return False
            items = self.getItemByName(item_name) or []
            items = [i for i in items if self.getTaurusParentItem(
                i) not in (items + self._selectedItems)]
            self.debug('In TaurusGraphicsScene.selectGraphicItem(%s)): matched %d items' % (
                item_name, len(items)))

        if self._selectionStyle == SynopticSelectionStyle.ELLIPSE:
            displaySelection = self._displaySelectionAsEllipse
        elif self._selectionStyle == SynopticSelectionStyle.OUTLINE:
            displaySelection = self._displaySelectionAsOutline
        else:
            raise Exception("Unexpected selectionStyle '%s'" %
                            SynopticSelectionStyle.whatis(self._selectionStyle))
        return displaySelection(items)

    def _displaySelectionAsEllipse(self, items):
        retval = False
        for item in items:
            try:
                if ((isinstance(item, TaurusGraphicsItem) and item.getExtensions().get('noSelect'))
                            or (item in self._selection)
                            # or (item in tangoGroup)
                        ):
                    continue
                x, y = item.x(), item.y()
                rect = item.boundingRect()
                srect = self.sceneRect()
                # 0 has to be excluded to check grouped element
                if not (0 < x <= self.sceneRect().width() and 0 < y <= srect.height()):
                    rx, ry = rect.topLeft().x(), rect.topLeft().y()
                    self.debug('\tposition not well mapped (%s,%s), using rect bound (%s,%s) instead' % (
                        x, y, rx, ry))
                    x, y = rx, ry  # If the object is in the corner it will be also 0
                w, h = rect.width(), rect.height()
                if x < 0 or y < 0:
                    self.debug('Cannot draw SelectionMark for %s(%s)(%s,%s) in a negative position (%f,%f)' % (
                        type(item).__name__, item._name, w, h, x, y))
                else:
                    if type(item) in (TaurusTextAttributeItem, TaurusTextStateItem) and isinstance(self.getSelectionMark(), Qt.QGraphicsPixmapItem):
                        x, y, w, h = x - 20, y, 20, 20
                    self.drawSelectionMark(x, y, w, h)
                    self.debug('> Moved the SelectionMark to item %s(%s)(%s,%s) at %f,%f' % (
                        type(item).__name__, item._name, w, h, x, y))
                if item not in self._selectedItems:
                    self._selectedItems.append(item)
                retval = True
            except Exception as e:
                self.warning('selectGraphicsItem(%s) failed! %s' %
                             (getattr(item, '_name', item), str(e)))
                self.warning(traceback.format_exc())
                # return False
        return retval

    def _displaySelectionAsOutline(self, items):
        def _outline(shapes):
            """"Compute the boolean union from a list of QGraphicsItem. """
            shape = None
            # TODO we can use a stack instead of recursivity
            for s in shapes:
                # TODO we should skip text and things like that
                if isinstance(s, TaurusGroupItem):
                    s = _outline(s.childItems())
                    if s is None:
                        continue

                s = s.shape()
                if shape is not None:
                    shape = shape.united(s)
                else:
                    shape = s

            if shape is None:
                return None

            return Qt.QGraphicsPathItem(shape)

        # TODO we can cache the outline instead of computing it again and again
        selectionShape = _outline(items)
        if selectionShape:
            # copy-paste from getSelectionMark
            color = Qt.QColor(Qt.Qt.blue)
            color.setAlphaF(.10)
            pen = Qt.QPen(Qt.Qt.SolidLine)
            pen.setWidth(4)
            pen.setColor(Qt.QColor(Qt.Qt.blue))
            selectionShape.setBrush(color)
            selectionShape.setPen(pen)

            for item in items:
                if item not in self._selectedItems:
                    self._selectedItems.append(item)

            # TODO i dont think this function work... or i dont know how...
            # self.setSelectionMark(picture=selectionShape)
            # ... Then do it it with hands...
            # copy-paste from drawSelectionMark
            self._selection.append(selectionShape)
            # It's better to add it hidden to avoid resizings
            selectionShape.hide()
            self.addItem(selectionShape)
            # Put on Top
            selectionShape.setZValue(9999)
            selectionShape.show()
            self.updateSceneViews()

            return True

        return False

    def clearSelection(self):
        #self.debug('In clearSelection([%d])'%len(self._selectedItems))
        for i in self._selection:
            i.hide()
            self.removeItem(i)
        self._selection = []
        self._selectedItems = []
        self.updateSceneViews()

    def setSelectionMark(self, picture=None, w=10, h=10):
        """This method allows to set a callable, graphic item or pixmap as selection mark (by default creates a blue circle).
        If picture is a callable, the object returned will be used as selection mark.
        If picture is a QGraphicsItem it will be used as selection mark.
        If picture is a QPixmap or a path to a pixmap a QGraphicsPixmapItem will be created.
        If no picture is provided, a blue ellipse will be drawn around the selected object.
        h/w will be used for height/width of the drawn object.
        """
        #self.debug('In setSelectionMark(%s,%d,%d)'%(picture,w,h))
        if picture is None:
            self.SelectionMark = None  # Reset of previous icon generators
        else:
            self.SelectionMark = (lambda p=picture, x=w,
                                  y=h: self.getSelectionMark(p, x, y))
        return self.SelectionMark

    def getSelectionMark(self, picture=None, w=10, h=10):
        if picture is None:
            if self.SelectionMark:
                SelectionMark = self.SelectionMark()
            else:
                SelectionMark = Qt.QGraphicsEllipseItem()
                color = Qt.QColor(Qt.Qt.blue)
                color.setAlphaF(.10)
                SelectionMark.setBrush(color)
                pen = Qt.QPen(Qt.Qt.CustomDashLine)
                pen.setWidth(4)
                pen.setColor(Qt.QColor(Qt.Qt.blue))
                SelectionMark.setPen(pen)
                SelectionMark.hide()  # It's better to add it hidden to avoid resizings
        else:
            try:
                if isinstance(picture, Qt.QGraphicsItem):
                    SelectionMark = picture
                    SelectionMark.setRect(0, 0, w, h)
                    SelectionMark.hide()
                elif hasattr(picture, '__call__'):
                    SelectionMark = picture()
                else:
                    if isinstance(picture, Qt.QPixmap):
                        pixmap = picture
                    elif isinstance(picture, string_types):
                        picture = str(picture)
                        pixmap = Qt.QPixmap(os.path.realpath(picture))
                    SelectionMark = Qt.QGraphicsPixmapItem()
                    SelectionMark.setPixmap(pixmap.scaled(w, h))
                    SelectionMark.hide()
            except:
                self.debug('In setSelectionMark(%s): %s' %
                           (picture, traceback.format_exc()))
                picture = None
        return SelectionMark

    def drawSelectionMark(self, x, y, w, h, oversize=1):
        """
        If h or w are None the mark is drawn at x,y
        If h or w has a value the mark is drawn in the center of the region ((x,y)(x+w,y+h))
        """
        #self.debug('%s has parent %s' % (item_name,getattr(item.parentItem(),'_name','ung')  if item.parentItem() else 'None'))
        #self.debug('drawSelectionMark(): center and width,height are: (%d,%d),(%d,%d)' % (x,y,w,h))

        mark = self.getSelectionMark()
        self._selection.append(mark)
        srect = self.itemsBoundingRect()
        MAX_CIRCLE_SIZE = srect.width(), srect.height()  # 500,500 #20,20
        LIMITS = (0, 0, srect.width(), srect.height())

        def bound(coords, bounds=LIMITS):
            """ x,y,w,h """
            x, y, w, h = coords
            if x < bounds[0]:
                w, x = w - (bounds[0] - x), bounds[0]
            if y < bounds[1]:
                h, y = h - (bounds[1] - y), bounds[1]
            if x + w > bounds[2]:
                w, x = (bounds[2] - x), x
            if y + h > bounds[3]:
                h, y = (bounds[3] - y), y
            return x, y, w, h

        if isinstance(mark, Qt.QGraphicsEllipseItem):
            if None not in [w, h]:
                if w > MAX_CIRCLE_SIZE[0] or h > MAX_CIRCLE_SIZE[1]:
                    # Applying correction if the file is too big, half max
                    # circle size around the center
                    x, y = (x + w / 2.) - .5 * \
                        MAX_CIRCLE_SIZE[0], (y + h / 2.) - .5 * \
                        MAX_CIRCLE_SIZE[1],
                    w, h = [.5 * t for t in MAX_CIRCLE_SIZE]
                else:
                    x, y = x - .5 * w, y - .5 * h
            else:
                w, h = [.5 * t for t in MAX_CIRCLE_SIZE]
                #self.debug('drawSelectionMark(): center and width,height are: (%d,%d),(%d,%d)' % (x,y,w,h))
            mark.setRect(*bound((x, y, w * 2, h * 2)))
            # mark.setRect(x,y,w*2,h*2)
        elif isinstance(mark, Qt.QGraphicsPixmapItem):
            rect = mark.boundingRect()
            if None not in [w, h]:
                x, y = x + .5 * w, y + .5 * h
            mark.setOffset(x - .5 * rect.width(), y - .5 * rect.height())
        elif isinstance(mark, Qt.QGraphicsItem):
            mark.setRect(x, y, w, h)

        mark.hide()  # It's better to add it hidden to avoid resizings
        self.addItem(mark)
        mark.setZValue(9999)  # Put on Top
        mark.show()
        self.updateSceneViews()
        return

    def getShellCommand(self, obj, wait=False):
        shellCom = obj.getExtensions().get('shellCommand').replace(
            '$NAME', obj._name).replace('$MODEL', obj._name)
        if not wait and not shellCom.endswith('&'):
            shellCom += ' &'
        if obj.noPrompt:
            subprocess.call(shellCom, shell=True)
        else:
            yes = Qt.QMessageBox.Ok
            no = Qt.QMessageBox.Cancel
            result = Qt.QMessageBox.question(self.parent(
            ), "Shell command", "Would you like to call shell command '" + shellCom + "' ?", yes, no)
            if result == yes:
                subprocess.call(shellCom, shell=True)
        return

    def getClass(self, clName):
        if not clName or clName == 'noPanel':
            return None
        elif clName in ('atkpanel.MainPanel', 'atkpanel'):
            clName = 'TaurusDevicePanel'
        # TODO: allow passing class names including module, e.g.: 'foo.Bar'
        if clName in globals():
            return globals()[clName]
        elif clName in locals():
            return locals()[clName]
        elif clName in dir(Qt):
            return getattr(Qt, clName)
        else:
            wf = TaurusWidgetFactory()
            try:
                return wf.getTaurusWidgetClass(clName)
            except:
                return None

    @staticmethod
    def getTaurusParentItem(item, top=True):
        """ Searches within a group hierarchy and returns a parent Taurus component or None if no parent TaurusBaseComponent
            is found."""
        if item is None:
            return None
        first, p = None, item.parentItem()
        while p:
            if isinstance(p, TaurusGraphicsItem):
                if first is None:
                    first = p
                    if not top:
                        break
                elif str(p.getModel()) != str(first.getModel()):
                    break
                else:
                    first = p
            p = p.parentItem()
        return first

    def getAllChildren(self, item, klass=None):
        """ Returns all children elements, filtering by klass if wanted """
        result = []
        try:
            children = item.childItems()
            result.extend(
                c for c in children if not klass or isinstance(c, klass))
            result.extend(c.childItems()
                          for c in children if not klass or isinstance(c, klass))
        except:
            pass
        return result

    def start(self):
        if self.updateThread:
            return
        self.updateQueue = Queue()
        self.updateThread = TaurusGraphicsUpdateThread(self)
        self.updateThread.start()  # Qt.QThread.HighPriority)

    def getQueue(self):
        return self.updateQueue

    def updateSceneItem(self, item):
        self.updateQueue.put(item)

    def updateSceneItems(self, items):
        self.updateQueue.put(items)

    def updateScene(self):
        self.update()

    def updateSceneViews(self):
        for v in self.views():
            v.viewport().update()
            # v.invalidateScene(self.SelectionCircle.boundingRect())
        return


class QGraphicsTextBoxing(Qt.QGraphicsItemGroup):
    """Display a text inside a virtual box. Support horizontal and vertical alignment"""

    _TEXT_RATIO = 0.8

    def __init__(self, parent=None, scene=None):
        Qt.QGraphicsItemGroup.__init__(self, parent)
        if scene is not None:
            scene.addItem(self)
        self._rect = Qt.QGraphicsRectItem(self)
        if scene is not None:
            scene.addItem(self._rect)
        self._rect.setBrush(Qt.QBrush(Qt.Qt.NoBrush))
        self._rect.setPen(Qt.QPen(Qt.Qt.NoPen))
        self._text = Qt.QGraphicsTextItem(self)
        if scene is not None:
            scene.addItem(self._text)
        self._text.setTransform(
            Qt.QTransform.fromScale(self._TEXT_RATIO, self._TEXT_RATIO),
            True)

        self._validBackground = None
        # using that like the previous code create a worst result
        self.__layoutValide = True
        self._alignment = Qt.Qt.AlignCenter | Qt.Qt.AlignVCenter

    def setRect(self, x, y, width, height):
        self._rect.setRect(x, y, width, height)
        self._invalidateLayout()

    def setPlainText(self, text):
        self._text.setPlainText(text)
        self._invalidateLayout()

    def setValidBackground(self, color):
        self._validBackground = color

    def toPlainText(self):
        return self._text.toPlainText()

    def brush(self):
        return self._rect.brush()

    def setBrush(self, brush):
        self._rect.setBrush(brush)

    def pen(self):
        return self._rect.pen()

    def setPen(self, pen):
        self._rect.setPen(pen)

    def setDefaultTextColor(self, color):
        self._text.setDefaultTextColor(color)

    def setHtml(self, html):
        self._text.setHtml(html)
        self._invalidateLayout()

    def setFont(self, font):
        self._text.setFont(font)
        self._invalidateLayout()

    def setAlignment(self, alignment):
        self._alignment = alignment
        self._invalidateLayout()

    def _invalidateLayout(self):
        """Invalidate the current location of the text"""
        if not self.__layoutValide:
            return
        self.__layoutValide = False
        self.update()

    def _validateLayout(self):
        """Compute the text location"""
        if self.__layoutValide:
            return

        rect = self._rect.rect()
        width, height = rect.width(), rect.height()
        textRect = self._text.boundingRect()

        # horizontal layout
        x = rect.x()
        alignment = int(self._alignment)
        if (alignment & int(Qt.Qt.AlignLeft)) != 0:
            x += 0
        elif (alignment & int(Qt.Qt.AlignHCenter)) != 0:
            x += width * 0.5 - textRect.width() * 0.5 * self._TEXT_RATIO
        elif (alignment & int(Qt.Qt.AlignRight)) != 0:
            x += width - textRect.width() * self._TEXT_RATIO

        # vertical layout
        y = rect.y()
        if (alignment & int(Qt.Qt.AlignTop)) != 0:
            y += 0
        elif (alignment & int(Qt.Qt.AlignVCenter)) != 0:
            y += height * 0.5 - textRect.height() * 0.5 * self._TEXT_RATIO
        elif (alignment & int(Qt.Qt.AlignBottom)) != 0:
            y += height - textRect.height() * self._TEXT_RATIO

        self._text.setPos(x, y)
        self.__layoutValide = True

    def paint(self, painter, option, widget):
        self._validateLayout()
        Qt.QGraphicsItemGroup.paint(self, painter, option, widget)


class QSpline(Qt.QGraphicsPathItem):

    def __init__(self, parent=None, closed=False, control_points=None):
        super(QSpline, self).__init__(parent)
        self.__closed = closed
        if control_points is None:
            control_points = []
        self.setControlPoints(control_points)

    def setControlPoints(self, control_points):
        self.__control_points = control_points
        self.updateSplinePath()

    def setClose(self, isClosed):
        if self.__closed == isClosed:
            return
        self.__closed = isClosed
        self.updateSplinePath()

    def updateSplinePath(self):
        path = Qt.QPainterPath()
        cp = self.__control_points
        nb_points = len(cp)
        if nb_points <= 1:
            pass
        elif nb_points == 2:
            path.moveTo(cp[0])
            path.lineTo(cp[1])
        else:
            path.moveTo(cp[0])
            for i in range(1, nb_points - 1, 3):
                p1 = cp[i + 0]
                p2 = cp[i + 1]
                end = cp[i + 2]
                path.cubicTo(p1, p2, end)
            if self.__closed:
                path.lineTo(cp[0])

        self.setPath(path)


class TaurusGraphicsItem(TaurusBaseComponent):
    """Base class for all Taurus Graphics Items"""

    def __init__(self, name=None, parent=None):
        self.call__init__(TaurusBaseComponent, name,
                          parent)  # <- log created here
        #self.debug('TaurusGraphicsItem(%s,%s)' % (name,parent))
        self.ignoreRepaint = False
        self.setName(name)
        self._currFgBrush = None
        self._currBgBrush = None
        self._currText = None
        self._currHtmlText = None
        self._map = None
        self._default = None
        self._visible = None
        # self.getExtensions() <= It must be called AFTER set_common_params()
        # in getGraphicsItem()
        self._contextMenu = []

    def setName(self, name):
        name = str(name or self.__class__.__name__)
        # srubio@cells.es: modified to store ._name since initialization (even
        # if a model is not set)
        self._name = name

    def getName(self):
        return self._name

    def setContextMenu(self, menu):
        '''Context Menu must be a list of tuples (ActionName,ActionMethod), empty tuples insert separators between options.'''
        self._contextMenu = menu

    def contextMenu(self):
        return self._contextMenu

    def getExtensions(self):
        """
        Any in
        ExtensionsList,noPrompt,standAlone,noTooltip,noSelect,ignoreRepaint,shellCommand,className,classParams
        """
        self._extensions = getattr(self, '_extensions', {})
        if 'ExtensionsList' in self._extensions:
            self._extensions.update((k.strip(), True) for k in self._extensions[
                                    'ExtensionsList'].split(','))
            self._extensions.pop('ExtensionsList')
        for k in ('noPrompt', 'standAlone', 'noTooltip', 'ignoreRepaint', 'noSelect'):
            if self._extensions.get(k, None) == '':
                self._extensions[k] = True
        self.noPrompt = self._extensions.get('noPrompt', False)
        self.standAlone = self._extensions.get('standAlone', False)
        self.noTooltip = self._extensions.get('noTooltip', False)
        self.ignoreRepaint = self._extensions.get(
            'ignoreRepaint', self.ignoreRepaint)
        self.setName(self._extensions.get('name', self._name))
        tooltip = '' if (self.noTooltip or self._name ==
                         self.__class__.__name__ or self._name is None) else str(self._name)
        #self.debug('setting %s.tooltip = %s'%(self._name,tooltip))
        self.setToolTip(tooltip)
        #self.debug('%s.getExtensions(): %s'%(self._name,self._extensions))
        return self._extensions

    #-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-
    # Mandatory methods to be implemented in any subclass of TaurusComponent
    #-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-

    def setModel(self, model):
        #self.debug('In %s.setModel(%s)'%(type(self).__name__,model))
        self.setName(model)

        if issubclass(Manager().findObjectClass(self._name), TaurusDevice):
            model = self._name + '/state'
        TaurusBaseComponent.setModel(self, model)

    def getParentTaurusComponent(self):
        """ Returns a parent Taurus component or None if no parent TaurusBaseComponent
            is found."""
        p = self.parentItem()
        while p and not isinstance(p, TaurusGraphicsItem):
            p = self.parentItem()
        return p

    # def fireEvent(self, type):
    def fireEvent(self, evt_src=None, evt_type=None, evt_value=None):
        """fires a value changed event to all listeners"""
        self.updateStyle()

    def updateStyle(self):
        """ Method called when the component detects an event that triggers a change
            in the style."""
        if self.scene():
            self.scene().updateSceneItem(self)

    # getDisplayValue is already (and better) implemented in TaurusBaseComponent
    # def getDisplayValue(self):
        #attrvalue = self.getModelValueObj()
        # if not attrvalue:
            # return self.getNoneValue()
        # return str(attrvalue.value)

    def isReadOnly(self):
        return True

    def __str__(self):
        return self.log_name + "(" + self.modelName + ")"

    def getModelClass(self):
        return TaurusAttribute


class TaurusGraphicsAttributeItem(TaurusGraphicsItem):
    """
    This class show value->text conversion in label widgets.
    Quality is shown in background
    """

    def __init__(self, name=None, parent=None):
        name = name or self.__class__.__name__
        self._unitVisible = True
        self._currValue = None
        self._userFormat = None
        self._unitVisible = True
        self.call__init__(TaurusGraphicsItem, name, parent)

    @deprecation_decorator(alt=".getDisplayValue(fragmentName='rvalue.units')",
                           rel="4.0.3")
    def getUnit(self):
        return self.getDisplayValue(fragmentName='rvalue.units')

    def updateStyle(self):
        v = self.getModelValueObj()
        if self.getShowQuality():
            try:
                quality = None
                if v:
                    quality = v.quality
                if quality == AttrQuality.ATTR_VALID and self._validBackground:
                    background = self._validBackground
                else:
                    background, _ = QT_ATTRIBUTE_QUALITY_PALETTE.qcolor(
                        quality)
                self.setBrush(Qt.QBrush(background))
            except:
                self.warning('In TaurusGraphicsAttributeItem(%s).updateStyle(%s): colors failed!' % (
                    self._name, self._currText))
                self.warning(traceback.format_exc())

        if v and self._userFormat:
            # TODO: consider extending this to use newer pyhon formatting syntax
            if hasattr(v.rvalue, 'magnitude'):
                text = self._userFormat % v.rvalue.magnitude
            else:
                text = self._userFormat % v.rvalue
            if self._unitVisible:
                text = "{0} {1.rvalue.units:~s}".format(text, v)
        else:
            if self._unitVisible:
                _frName = None
            else:
                _frName = 'rvalue.magnitude'
            text = self.getDisplayValue(fragmentName=_frName)

        self._currText = text
        self._currHtmlText = None

        TaurusGraphicsItem.updateStyle(self)

    def setUserFormat(self, format):
        self._userFormat = format

    def setUnitVisible(self, yesno):
        self._unitVisible = yesno


class TaurusGraphicsStateItem(TaurusGraphicsItem):
    """
    In State Item the displayValue should not override the label
    This item will modify only foreground/background colors
    """

    def __init__(self, name=None, parent=None):
        name = name or self.__class__.__name__
        self.call__init__(TaurusGraphicsItem, name, parent)

    def updateStyle(self):

        from taurus.core.tango import DevState  # Tango-centric

        v = self.getModelValueObj()

        self._currBrush = Qt.QBrush(Qt.Qt.NoBrush)
        if v:  # or self.getShowState():
            try:
                bg_brush, fg_brush = None, None
                if self.getModelObj().getType() == DataType.DevState:
                    bg_brush, fg_brush = QT_DEVICE_STATE_PALETTE.qbrush(
                        v.rvalue)
                elif self.getModelObj().getType() == DataType.Boolean:
                    bg_brush, fg_brush = QT_DEVICE_STATE_PALETTE.qbrush(
                        (DevState.FAULT, DevState.ON)[v.rvalue])
                elif self.getShowQuality():
                    bg_brush, fg_brush = QT_ATTRIBUTE_QUALITY_PALETTE.qbrush(
                        v.quality)
                if None not in (bg_brush, fg_brush):
                    self._currBgBrush = bg_brush
                    self._currFgBrush = fg_brush
                    # If there's no filling, applying background brush to
                    # foreground
                    if Qt.Qt.NoBrush != getattr(self, '_fillStyle', Qt.Qt.NoBrush):
                        #self.debug('In TaurusGraphicsStateItem(%s).updateStyle(%s): switching background to foreground'%(self._name,self._currText))
                        self._currFgBrush = bg_brush
                    if self._currText:
                        self._currHtmlText = '<p style="color:%s">%s</p>' % (
                            self._currBgBrush.color().name(), self._currText)
            except:
                self.warning('In TaurusGraphicsStateItem(%s).updateStyle(%s): colors failed!' % (
                    self._name, self._currText))
                self.warning(traceback.format_exc())

        # Parsing _map to manage visibility (a list of values for which the
        # item is visible or not)
        if (v and not self._map is None and
                self._currText in DevState.__members__):
            #self.debug('In TaurusGraphicsStateItem.updateStyle(): mapping %s'%self._currText)
            if DevState[self._currText] == self._map[1]:
                self.setVisible(self._map[2])
                self._visible = self._map[2]
            else:
                self.setVisible(self._default)
                self._visible = self._default

        TaurusGraphicsItem.updateStyle(self)


class TaurusEllipseStateItem(Qt.QGraphicsEllipseItem, TaurusGraphicsStateItem):

    def __init__(self, name=None, parent=None, scene=None):
        name = name or self.__class__.__name__
        Qt.QGraphicsEllipseItem.__init__(self, parent)
        if scene is not None:
            scene.addItem(self)
        self.call__init__(TaurusGraphicsStateItem, name, parent)

    def paint(self, painter, option, widget=None):
        if self._currBgBrush:
            self._currBgBrush.setStyle(self.brush().style())
            self.setBrush(self._currBgBrush)
        Qt.QGraphicsEllipseItem.paint(self, painter, option, widget)


class TaurusRectStateItem(Qt.QGraphicsRectItem, TaurusGraphicsStateItem):

    def __init__(self, name=None, parent=None, scene=None):
        name = name or self.__class__.__name__
        Qt.QGraphicsRectItem.__init__(self, parent)
        if scene is not None:
            scene.addItem(self)
        self.call__init__(TaurusGraphicsStateItem, name, parent)

    def paint(self, painter, option, widget):
        if self._currBgBrush:
            self._currBgBrush.setStyle(self.brush().style())
            self.setBrush(self._currBgBrush)
        Qt.QGraphicsRectItem.paint(self, painter, option, widget)


class TaurusSplineStateItem(QSpline, TaurusGraphicsStateItem):

    def __init__(self, name=None, parent=None, scene=None):
        name = name or self.__class__.__name__
        QSpline.__init__(self, parent)
        if scene is not None:
            scene.addItem(self)
        self.call__init__(TaurusGraphicsStateItem, name, parent)

    def paint(self, painter, option, widget):
        if self._currBgBrush:
            self._currBgBrush.setStyle(self.brush().style())
            self.setBrush(self._currBgBrush)
        QSpline.paint(self, painter, option, widget)


class TaurusRoundRectItem(Qt.QGraphicsPathItem):

    def __init__(self, name=None, parent=None, scene=None):
        Qt.QGraphicsPathItem.__init__(self, parent)
        if scene is not None:
            scene.addItem(self)
        self.__rect = None
        self.setCornerWidth(0, 0)

    def __updatePath(self):
        if self.__rect is None:
            return
        if self.__corner is None:
            return

        path = Qt.QPainterPath()
        cornerWidth, nbPoints = self.__corner
        if cornerWidth == 0 or nbPoints == 0:
            path.addRect(self.__rect)
        elif cornerWidth * 2 > self.__rect.width():
            path.addRect(self.__rect)
        elif cornerWidth * 2 > self.__rect.height():
            path.addRect(self.__rect)
        else:
            path.addRoundedRect(self.__rect, cornerWidth, cornerWidth)
        self.setPath(path)

    def setRect(self, x, y, width, height):
        self.__rect = Qt.QRectF(x, y, width, height)
        self.__updatePath()

    def setCornerWidth(self, width, nbPoints):
        self.__corner = width, nbPoints
        self.__updatePath()


class TaurusRoundRectStateItem(TaurusRoundRectItem, TaurusGraphicsStateItem):

    def __init__(self, name=None, parent=None, scene=None):
        name = name or self.__class__.__name__
        TaurusRoundRectItem.__init__(self, parent, scene)
        self.call__init__(TaurusGraphicsStateItem, name, parent)

    def paint(self, painter, option, widget):
        if self._currBgBrush:
            self._currBgBrush.setStyle(self.brush().style())
            self.setBrush(self._currBgBrush)
        TaurusRoundRectItem.paint(self, painter, option, widget)


class TaurusGroupItem(Qt.QGraphicsItemGroup):

    def __init__(self, name=None, parent=None, scene=None):
        Qt.QGraphicsItemGroup.__init__(self, parent)
        if scene is not None:
            scene.addItem(self)


class TaurusGroupStateItem(TaurusGroupItem, TaurusGraphicsStateItem):

    def __init__(self, name=None, parent=None, scene=None):
        name = name or self.__class__.__name__
        TaurusGroupItem.__init__(self, parent, scene)
        self.call__init__(TaurusGraphicsStateItem, name, parent)

    def paint(self, painter, option, widget):
        TaurusGroupItem.paint(self, painter, option, widget)


class TaurusPolygonStateItem(Qt.QGraphicsPolygonItem, TaurusGraphicsStateItem):

    def __init__(self, name=None, parent=None, scene=None):
        name = name or self.__class__.__name__
        #Qt.QGraphicsRectItem.__init__(self, parent)
        Qt.QGraphicsPolygonItem.__init__(self, parent)
        if scene is not None:
            scene.addItem(self)
        self.call__init__(TaurusGraphicsStateItem, name, parent)

    def paint(self, painter, option, widget):
        if self._currBgBrush:
            self._currBgBrush.setStyle(self.brush().style())
            self.setBrush(self._currBgBrush)
        Qt.QGraphicsPolygonItem.paint(self, painter, option, widget)


class TaurusLineStateItem(Qt.QGraphicsLineItem, TaurusGraphicsStateItem):

    def __init__(self, name=None, parent=None, scene=None):
        name = name or self.__class__.__name__
        Qt.QGraphicsLineItem.__init__(self, parent)
        if scene is not None:
            scene.addItem(self)
        self.call__init__(TaurusGraphicsStateItem, name, parent)

    def paint(self, painter, option, widget):
        if self._currBgBrush:
            self._currBgBrush.setStyle(self.brush().style())
            self.setBrush(self._currBgBrush)
        Qt.QGraphicsLineItem.paint(self, painter, option, widget)


class TaurusTextStateItem(QGraphicsTextBoxing, TaurusGraphicsStateItem):
    """
    A QGraphicsItem that represents a text related to a device state or attribute quality
    """

    def __init__(self, name=None, parent=None, scene=None):
        name = name or self.__class__.__name__
        QGraphicsTextBoxing.__init__(self, parent, scene)
        self.call__init__(TaurusGraphicsStateItem, name, parent)

    def paint(self, painter, option, widget):
        if self._currHtmlText:
            self.setHtml(self._currHtmlText)
        else:
            self.setPlainText(self._currText or '')
        QGraphicsTextBoxing.paint(self, painter, option, widget)


class TaurusTextAttributeItem(QGraphicsTextBoxing, TaurusGraphicsAttributeItem):
    """
    A QGraphicsItem that represents a text related to an attribute value
    """

    def __init__(self, name=None, parent=None, scene=None):
        name = name or self.__class__.__name__
        QGraphicsTextBoxing.__init__(self, parent, scene)
        self.call__init__(TaurusGraphicsAttributeItem, name, parent)

    def paint(self, painter, option, widget):
        # self.debug('TaurusTextAttributeItem(%s,%s,%s)'%(self.getName(),self._currText,self._currHtmlText))
        if self._currHtmlText:
            self.setHtml(self._currHtmlText)
        else:
            self.setPlainText(self._currText or '')
        QGraphicsTextBoxing.paint(self, painter, option, widget)


TYPE_TO_GRAPHICS = {
    None: {"Rectangle": Qt.QGraphicsRectItem,
           "RoundRectangle": TaurusRoundRectItem,
           "Ellipse": Qt.QGraphicsEllipseItem,
           "Polyline": Qt.QGraphicsPolygonItem,
           "Label": QGraphicsTextBoxing,
           "Line": Qt.QGraphicsLineItem,
           "Group": TaurusGroupItem,
           "SwingObject": TaurusTextAttributeItem,
           "Image": Qt.QGraphicsPixmapItem,
           "Spline": QSpline, },

    TaurusDevice: {"Rectangle": TaurusRectStateItem,
                   "RoundRectangle": TaurusRoundRectStateItem,
                   "Ellipse": TaurusEllipseStateItem,
                   "Polyline": TaurusPolygonStateItem,
                   "Label": TaurusTextStateItem,
                   "Line": Qt.QGraphicsLineItem,  # TaurusLineStateItem,
                   "Group": TaurusGroupStateItem,
                   "SwingObject": TaurusTextAttributeItem,
                   "Image": Qt.QGraphicsPixmapItem,
                   "Spline": TaurusSplineStateItem, },

    TaurusAttribute: {"Rectangle": TaurusRectStateItem,
                      "RoundRectangle": TaurusRoundRectStateItem,
                      "Ellipse": TaurusEllipseStateItem,
                      "Polyline": TaurusPolygonStateItem,
                      "Label": TaurusTextAttributeItem,
                      "Line": Qt.QGraphicsLineItem,  # TaurusLineStateItem,
                      "Group": TaurusGroupStateItem,
                      "SwingObject": TaurusTextAttributeItem,
                      "Image": Qt.QGraphicsPixmapItem,
                      "Spline": TaurusSplineStateItem, },
}


class TaurusBaseGraphicsFactory(object):

    def __init__(self):
        pass

    def getSceneObj(self):
        raise RuntimeError(
            "Invalid call to AbstractGraphicsFactory::getSceneObj()")

    def getObj(self, name, params):
        raise RuntimeError("Invalid call to AbstractGraphicsFactory::getObj()")

    def getRectangleObj(self, params):
        raise RuntimeError(
            "Invalid call to AbstractGraphicsFactory::getRectangleObj()")

    def getRoundRectangleObj(self, params):
        raise RuntimeError(
            "Invalid call to AbstractGraphicsFactory::getRoundRectangleObj()")

    def getLineObj(self, params):
        raise RuntimeError(
            "Invalid call to AbstractGraphicsFactory::getLineObj()")

    def getEllipseObj(self, params):
        raise RuntimeError(
            "Invalid call to AbstractGraphicsFactory::getEllipseObj()")

    def getPolylineObj(self, params):
        raise RuntimeError(
            "Invalid call to AbstractGraphicsFactory::getPolylineObj()")

    def getLabelObj(self, params):
        raise RuntimeError(
            "Invalid call to AbstractGraphicsFactory::getLabelObj()")

    def getGroupObj(self, params):
        raise RuntimeError(
            "Invalid call to AbstractGraphicsFactory::getGroupObj()")

    def getSwingObjectObj(self, params):
        raise RuntimeError(
            "Invalid call to AbstractGraphicsFactory::getSwingObjectObj()")

    def getImageObj(self, parms):
        raise RuntimeError(
            "Invalid call to AbstractGraphicsFactory::getImageObj()")

    def getSplineObj(self, params):
        raise RuntimeError(
            "Invalid call to AbstractGraphicsFactory::getSplineObj()")

    def getGraphicsClassItem(self, cls, type_):
        ncls = cls
        try:
            if issubclass(cls, TaurusDevice):
                ncls = TaurusDevice
            elif issubclass(cls, TaurusAttribute):
                ncls = TaurusAttribute
        except:
            pass
        ncls = TYPE_TO_GRAPHICS.get(
            ncls, TYPE_TO_GRAPHICS.get(None)).get(type_)
        return ncls

    def getGraphicsItem(self, type_, params):
        name = params.get(self.getNameParam())
        # applying alias
        for k, v in getattr(self, 'alias', {}).items():
            if k in name:
                name = str(name).replace(k, v)
                params[self.getNameParam()] = name
        cls = None
        # TODO: starting slashes are allowed while we support parent model
        #       feature (taurus-org/taurus#734)
        if not name.startswith("/") and "/" in name:
            try:
                from taurus.core.tango.tangovalidator import (
                    TangoDeviceNameValidator,
                    TangoAttributeNameValidator,
                )
            except ImportError:
                pass
            else:
                if (TangoDeviceNameValidator().isValid(name)
                        or TangoAttributeNameValidator().isValid(name)):
                    # replacing Taco identifiers in %s'%name
                    if name.lower().startswith("tango:"):
                        if name.count('/') == 2 or 'tango:/' not in name.lower():  # noqa
                            nname = name.split(':', 1)[-1]
                            params[self.getNameParam()] = name = nname
                    else:
                        from taurus import warning
                        warning("if you use a tango name as JD name it must "
                                "with tango:")
                    if name.lower().endswith('/state'):
                        name = name.rsplit('/', 1)[0]
                    cls = Manager().findObjectClass(name)
        else:
            if name:
                self.debug('%s does not match a tango name' % name)
        klass = self.getGraphicsClassItem(cls, type_)
        item = klass()
        # It's here were Attributes are subscribed
        self.set_common_params(item, params)
        if hasattr(item, 'getExtensions'):
            item.getExtensions()  # <= must be called here to take extensions from params
        return item

    def getNameParam(self):
        """Returns the name of the parameter which contains the name identifier.
           Default implementation returns 'name'.
           Overwrite has necessary."""
        return 'name'

    def set_common_params(self, item, params):
        """Sets the common parameters. Default implementation does nothing.
           Overwrite has necessary."""
        pass



#!/usr/bin/env python

#############################################################################
##
# This file is part of Taurus
##
# http://taurus-scada.org
##
# Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
# Taurus is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
##
# Taurus is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
##
# You should have received a copy of the GNU Lesser General Public License
# along with Taurus.  If not, see <http://www.gnu.org/licenses/>.
##
#############################################################################
"""
This module contains provides the fqdn_no_alias function.
"""

import socket
from taurus import warning


def fqdn_no_alias(hostname):
    """
    Normalize hostname so that it is a full name (including domain name).

    .. note:: this is similar to socket.getfqdn, but it avoids translating
              aliases into their "real" address.
    """
    if hostname == "localhost" or "." in hostname:
        # optimization: leave localhost or names including domain as they are
        return hostname
    try:
        real, aliases, _ = socket.gethostbyname_ex(hostname)
    except (socket.gaierror, socket.herror) as e:
        # return the given hostname in case of error
        return hostname

    if aliases:
        if hostname == real or hostname in aliases:
            # in some corner cases (e.g. when defining several aliases in a
            # hosts file), the hostname may be one of the aliases.
            return hostname

        # return alias if it exists
        if len(aliases) > 1:
            # warn in the (unusual) case that there is more than 1 alias
            warning("fqdn_no_alias: %s has %d aliases. Using the first one",
                    hostname,
                    len(aliases)
                    )
        return aliases[0]
    else:
        # if there are no aliases, return real
        return real


if __name__ == '__main__':

    # TODO: once we move to PyTest, test_fqdn_no_alias could be put in the
    #       official tests (test_fqdn_no_alias_local would require to be
    #       generalized)

    def test_fqdn_no_alias():
        """tests for test_fqdn_no_alias"""
        assert fqdn_no_alias("NONEXISTENT") == "NONEXISTENT"
        # test localhost
        assert fqdn_no_alias("localhost") == "localhost"
        # test with an aliased full hostname+domain (alias should not resolv!)
        print(socket.gethostbyname_ex('mail.google.com'))
        print("socket.getfqdn:", socket.getfqdn("mail.google.com"))
        assert fqdn_no_alias("mail.google.com") == "mail.google.com"


    def test_fqdn_no_alias_local(name="www"):
        import re
        # test with an existent host in your domain
        print(socket.gethostbyname_ex(name))
        print("socket.getfqdn:", socket.getfqdn(name))
        assert re.match(name + r"\.[a-zA-Z0-9\.-]+", fqdn_no_alias(name))


    test_fqdn_no_alias()
    test_fqdn_no_alias_local()






# -*- coding: utf-8 -*-

##############################################################################
##
## This file is part of Taurus
##
## http://taurus-scada.org
##
## Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
## Taurus is free software: you can redistribute it and/or modify
## it under the terms of the GNU Lesser General Public License as published by
## the Free Software Foundation, either version 3 of the License, or
## (at your option) any later version.
##
## Taurus is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU Lesser General Public License for more details.
##
## You should have received a copy of the GNU Lesser General Public License
## along with Taurus.  If not, see <http://www.gnu.org/licenses/>.
##
##############################################################################

"""This module exposes QtWebKit module"""

from . import PYQT5, PYQT4, PYSIDE, PYSIDE2, PythonQtError


if PYQT5:
    from PyQt5.QtWebKit import *
    # import * from QtWebkitWidgets for PyQt4 style compat
    from PyQt5.QtWebKitWidgets import *
elif PYSIDE2:
    from PySide2.QtWebKit import *
    # import * from QtWebkitWidgets for PyQt4 style compat
    from PySide2.QtWebKitWidgets import *
elif PYQT4:
    from PyQt4.QtWebKit import *
elif PYSIDE:
    from PySide.QtWebKit import *
else:
    raise PythonQtError('No Qt bindings could be found')



#!/usr/bin/env python

#############################################################################
##
# This file is part of Taurus
##
# http://taurus-scada.org
##
# Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
# Taurus is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
##
# Taurus is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
##
# You should have received a copy of the GNU Lesser General Public License
# along with Taurus.  If not, see <http://www.gnu.org/licenses/>.
##
#############################################################################

"""Module containing base classes for using the TangoSchemeTest DS in tests"""

from builtins import object
import PyTango
from taurus.core.tango.starter import ProcessStarter
from taurus.test import getResourcePath
import pytest
from random import randint


__all__ = ['TangoSchemeTestLauncher', 'taurus_test_ds']

__docformat__ = 'restructuredtext'


@pytest.fixture(scope="module")
def taurus_test_ds():
    """
    A pytest fixture that launches TangoSchemeTest for the test
    It provides the device name as the fixture value.

    Usage::
        from taurus.core.tango.test import taurus_test_ds

        def test_foo(taurus_test_ds):
            import taurus
            d = taurus.Device(taurus_test_ds)
            assert d["string_scalar"].rvalue == "hello world"

    """
    ds_name = 'TangoSchemeTest/unittest/temp-{:08d}'.format(
        randint(0, 99999999))

    # get path to DS and executable
    device = getResourcePath(
        'taurus.core.tango.test.res', 'TangoSchemeTest')
    # create starter for the device server
    _starter = ProcessStarter(device, 'TangoSchemeTest/unittest')
    # register
    _starter.addNewDevice(ds_name, klass='TangoSchemeTest')
    # start device server
    _starter.startDs()

    yield ds_name
    d = PyTango.DeviceProxy(ds_name)
    d.Reset()
    _starter.stopDs(hard_kill=True)
    # remove server
    _starter.cleanDb(force=True)


class TangoSchemeTestLauncher(object):
    """A base class for TestCase classes wishing to start a TangoSchemeTest.
    Use it as a mixin class"""

    DEV_NAME = 'TangoSchemeTest/unittest/temp-1'  # the dev name

    @classmethod
    def setUpClass(cls):
        """ Create and run a TangoSchemeTest device server
        """
        # get path to DS and executable
        device = getResourcePath('taurus.core.tango.test.res',
                                 'TangoSchemeTest')
        # create starter for the device server
        cls._starter = ProcessStarter(device, 'TangoSchemeTest/unittest')
        # register
        cls._starter.addNewDevice(cls.DEV_NAME, klass='TangoSchemeTest')
        # start device server
        cls._starter.startDs()

    @classmethod
    def tearDownClass(cls):
        """ Stop the device server and undo changes to the database
        """
        d = PyTango.DeviceProxy(cls.DEV_NAME)
        d.Reset()
        cls._starter.stopDs(hard_kill=True)
        # remove server
        cls._starter.cleanDb(force=True)

    def tearDown(self):
        d = PyTango.DeviceProxy(self.DEV_NAME)
        d.Reset()

if __name__ == '__main__':
    pass



#!/usr/bin/env python

#############################################################################
##
# This file is part of Taurus
##
# http://taurus-scada.org
##
# Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
# Taurus is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
##
# Taurus is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
##
# You should have received a copy of the GNU Lesser General Public License
# along with Taurus.  If not, see <http://www.gnu.org/licenses/>.
##
#############################################################################

""" Every TaurusWidget should have the following Qt Designer extended capabilities:

  - Task menu:
    it means when you right click on the widget in the designer, it will have
    the following additional items:
    - 'Edit model...' - opens a customized dialog for editing the widget model

  - Property Sheet:
    it means that in the Qt Designer property sheet it will have the following
    properties customized:
    - 'model' - will have a '...' button that will open a customized dialog for
      editing the widget model (same has 'Edit model...' task menu item
"""

from __future__ import print_function
import inspect

from taurus.external.qt import Qt
from taurus.external.qt import QtDesigner

from taurus.core.util.log import Logger


def Q_TYPEID(class_name):
    """ Helper function to generate an IID for Qt."""
    return "com.trolltech.Qt.Designer.%s" % class_name

designer_logger = Logger("PyQtDesigner")


class TaurusWidgetPlugin(QtDesigner.QPyDesignerCustomWidgetPlugin):
    """TaurusWidgetPlugin"""

    def __init__(self, parent=None):
        QtDesigner.QPyDesignerCustomWidgetPlugin.__init__(self)
        self._log = Logger(self._getWidgetClassName(), designer_logger)
        self.initialized = False

    def initialize(self, formEditor):
        """ Overwrite if necessary. Don't forget to call this method in case you
            want the generic taurus extensions in your widget."""
        if self.isInitialized():
            return

        self.initialized = True

    def isInitialized(self):
        return self.initialized

    def getWidgetClass(self):
        return self.WidgetClass

    def _getWidgetClassName(self):
        return self.getWidgetClass().__name__

    def __getWidgetArgs(self, klass=None, designMode=True, parent=None):
        if klass is None:
            klass = self.getWidgetClass()
        ctor = klass.__init__
        aspec = inspect.getargspec(ctor)
        if aspec.defaults is None:
            kwspec = {}
        else:
            kwspec = dict(zip(aspec.args[-len(aspec.defaults):],
                              aspec.defaults))
        args, kwargs = [], {}
        if 'designMode' in kwspec:
            kwargs['designMode'] = designMode
        if 'parent' in kwspec:
            kwargs['parent'] = parent
        else:
            args.append(parent)
        return args, kwargs

    def createWidget(self, parent):
        try:
            klass = self.getWidgetClass()
            args, kwargs = self.__getWidgetArgs(klass=klass,
                                                designMode=True,
                                                parent=parent)
            w = klass(*args, **kwargs)
        except Exception as e:
            name = self._getWidgetClassName()
            print(100 * "=")
            print("taurus designer plugin error creating %s: %s" % (name, str(e)))
            print(100 * "-")
            import traceback
            traceback.print_exc()
            w = None
        return w

    def getWidgetInfo(self, key, dft=None):
        if not hasattr(self, '_widgetInfo'):
            self._widgetInfo = self.getWidgetClass().getQtDesignerPluginInfo()
        return self._widgetInfo.get(key, dft)

    # This method returns the name of the custom widget class that is provided
    # by this plugin.
    def name(self):
        return self._getWidgetClassName()

    def group(self):
        """ Returns the name of the group in Qt Designer's widget box that this
            widget belongs to.
            It returns 'Taurus Widgets'. Overwrite if want another group."""
        return self.getWidgetInfo('group', 'Taurus Widgets')

    def getIconName(self):
        return self.getWidgetInfo('icon')

    def icon(self):
        icon = self.getWidgetInfo('icon')
        if icon is None:
            return Qt.QIcon()
        elif isinstance(icon, Qt.QIcon):
            return icon
        else:
            if icon.find(":") == -1:
                icon = 'designer:%s' % icon
            return Qt.QIcon(icon)

    def domXml(self):
        name = str(self.name())
        lowerName = name[0].lower() + name[1:]
        return '<widget class="%s" name=\"%s\" />\n' % (name, lowerName)

    def includeFile(self):
        """Returns the module containing the custom widget class. It may include
           a module path."""
        return self.getWidgetInfo('module')

    def toolTip(self):
        tooltip = self.getWidgetInfo('tooltip')
        if tooltip is None:
            tooltip = "A %s" % self._getWidgetClassName()
        return tooltip

    def whatsThis(self):
        whatsthis = self.getWidgetInfo('whatsthis')
        if whatsthis is None:
            whatsthis = "This is a %s widget" % self._getWidgetClassName()
        return whatsthis

    def isContainer(self):
        return self.getWidgetInfo('container', False)



import sys
from taurus.qt.qtgui.panel import TaurusForm
from taurus.qt.qtgui.display import TaurusLabel
from taurus.qt.qtgui.application import TaurusApplication

app = TaurusApplication(sys.argv, cmd_line_parser=None)

panel = TaurusForm()
props = ['state', 'status', 'position', 'velocity', 'acceleration']
model = ['sys/taurustest/1/%s' % p for p in props]
panel.setModel(model)
panel[0].readWidgetClass = TaurusLabel
panel[2].writeWidgetClass = 'TaurusWheelEdit'

panel.show()
sys.exit(app.exec_())



#!/usr/bin/env python

#############################################################################
##
# This file is part of Taurus
##
# http://taurus-scada.org
##
# Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
# Taurus is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
##
# Taurus is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
##
# You should have received a copy of the GNU Lesser General Public License
# along with Taurus.  If not, see <http://www.gnu.org/licenses/>.
##
#############################################################################

"""
Every TaurusWidget should have the following Qt Designer extended
capabilities:

  - Task menu:
    it means when you right click on the widget in the designer, it will have
    the following additional items:

    - 'Edit model...': opens a customized dialog for editing the widget model

  - Property Sheet:
    it means that in the Qt Designer property sheet it will have the following
    properties customized:

    - 'model': will have a '...' button that will open a customized dialog for
      editing the widget model (same has 'Edit model...' task menu item
"""

from __future__ import absolute_import

from taurus.external.qt import QtDesigner
from taurus.qt.qtgui.container.qcontainer import QGroupWidget


Q_TYPEID = {'QPyDesignerContainerExtension':     'com.trolltech.Qt.Designer.Container',
            'QPyDesignerPropertySheetExtension': 'com.trolltech.Qt.Designer.PropertySheet',
            'QPyDesignerTaskMenuExtension':      'com.trolltech.Qt.Designer.TaskMenu',
            'QPyDesignerMemberSheetExtension':   'com.trolltech.Qt.Designer.MemberSheet'}


class QGroupWidgetContainerExtension(QtDesigner.QPyDesignerContainerExtension):

    def __init__(self, widget, parent=None):
        super(QGroupWidgetContainerExtension, self).__init__(parent)
        self._widget = widget
        self._page_widget = None

    def addWidget(self, widget):
        if self.count() > 0:
            raise Exception("Can only have at maximum one child")
        self._layout().addWidget(widget)
        self._page_widget = widget

    def _content(self):
        return self._widget.content()

    def _layout(self):
        return self._content().layout()

    def count(self):
        return self._layout().count()

    def currentIndex(self):
        if self.count() > 0:
            return 0
        return -1

    def insertWidget(self, index, widget):
        self.addWidget(widget)

    def remove(self, index):
        self._layout().removeWidget(self.widget(index))

    def setCurrentIndex(self, index):
        pass

    def widget(self, index):
        return self._page_widget


class QGroupWidgetExtensionFactory(QtDesigner.QExtensionFactory):

    def __init__(self, parent=None):
        super(QGroupWidgetExtensionFactory, self).__init__(parent)

    def createExtension(self, obj, iid, parent):
        if iid != Q_TYPEID['QPyDesignerContainerExtension']:
            return None
        if isinstance(obj, QGroupWidget):
            return QGroupWidgetContainerExtension(obj, parent)
        return None


def create_plugin():
    from taurus.qt.qtdesigner.taurusplugin.taurusplugin import (
        TaurusWidgetPlugin)

    class QGroupWidgetPlugin(TaurusWidgetPlugin):

        WidgetClass = QGroupWidget

        def initialize(self, formEditor):
            if self.isInitialized():
                return

            manager = formEditor.extensionManager()
            if manager:
                self.factory = QGroupWidgetExtensionFactory(manager)
                manager.registerExtensions(self.factory, Q_TYPEID[
                                           'QPyDesignerContainerExtension'])
            self.initialized = True
    return QGroupWidgetPlugin

QGroupWidgetPlugin = create_plugin()



#!/usr/bin/env python

#############################################################################
##
# This file is part of Taurus
##
# http://taurus-scada.org
##
# Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
# Taurus is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
##
# Taurus is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
##
# You should have received a copy of the GNU Lesser General Public License
# along with Taurus.  If not, see <http://www.gnu.org/licenses/>.
##
#############################################################################

"""
comunications.py:
"""

from __future__ import print_function

from taurus.external.qt import QtCore, compat

import weakref

_DEBUG = False


def get_signal(obj, signalname):
    """Return signal from object and signal name."""
    if '(' not in signalname:
        return getattr(obj, signalname)
    name, dtype = signalname.strip(')').split('(')
    dtype = tuple(dtype.split(','))
    return getattr(obj, name)[dtype]


class DataModel(QtCore.QObject):
    '''
    An object containing one piece of data which is intended to be shared. The
    data will be identified by its UID (a unique identifier known to objects
    that intend to access the data)

    In general, you are not supposed to instantiate objects of this class
    directly. Instead, you should interact via the :class:`SharedDataManager`,
    which uses :meth:`SharedDataManager.__getDataModel` to ensure that the
    DataModels are singletons.
    '''

    dataChanged = QtCore.pyqtSignal(compat.PY_OBJECT)

    def __init__(self, parent, dataUID, defaultData=None):
        '''
        creator
        :param parent: (QObject) the object's parent
        :param dataUID: (str) a unique identifier for the Data Model
        '''
        QtCore.QObject.__init__(self, parent)
        self.__dataUID = dataUID
        self.__data = defaultData
        self.__isDataSet = False

        self.__readerSlots = []
        self.__writerSignals = []

    def __repr__(self):
        return '<DataModel object with dataUID="%s">' % self.dataUID()

    def dataUID(self):
        '''
        returns the data unique identifier

        :return: (str)
        '''
        return self.__dataUID

    def getData(self):
        '''
        Returns the data object.

        :return: (object) the data object
        '''
        return self.__data

    def setData(self, data):
        '''
        sets the data object and emits a "dataChanged" signal with the data as the parameter

        :param data: (object) the new value for the Model's data
        '''
        self.__data = data
        self.__isDataSet = True
        self.dataChanged.emit(self.__data)

    def connectReader(self, slot, readOnConnect=True):
        '''
        Registers the given slot method to receive notifications whenever the
        data is changed.

        :param slot: (callable) a method that will be called when the data changes.
                     This slot will be the receiver of a signal which has the
                     data as its first argument.
        :param readOnConnect: (bool) if True (default) the slot will be called
                              immediately with the current value of the data
                              if the data has been already initialized

        .. seealso:: :meth:`connectWriter`, :meth:`getData`
        '''
        self.dataChanged.connect(slot)

        if readOnConnect and self.__isDataSet:
            slot(self.__data)
        obj = getattr(slot, '__self__', slot)
        self.__readerSlots.append((weakref.ref(obj), slot.__name__))

    def connectWriter(self, writer, signalname):
        '''
        Registers the given writer object as a writer of the data. The writer is
        then expected to emit a `QtCore.SIGNAL(signalname)` with the new data as the
        first parameter.

        :param writer: (QObject) object that will change the data
        :param signalname: (str) the signal name that will notify changes
                           of the data

        .. seealso:: :meth:`connectReader`, :meth:`setData`
        '''
        try:
            get_signal(writer, signalname).connect(self.setData)
        except AttributeError:
            # support old-style signal
            self.connect(writer, QtCore.SIGNAL(signalname), self.setData)
        self.__writerSignals.append((weakref.ref(writer), signalname))

    def disconnectWriter(self, writer, signalname):
        '''unregister a writer from this data model

        :param writer: (QObject) object to unregister
        :param signalname: (str) the signal that was registered

        .. seealso:: :meth:`SharedDataManager.disconnectWriter`
        '''
        ok = get_signal(writer, signalname).disconnect(self.setData)
        self.__writerSignals.remove((weakref.ref(writer), signalname))

    def disconnectReader(self, slot):
        '''
        unregister a reader

        :param slot: (callable) the slot to which this was connected

        .. seealso:: :meth:`SharedDataManager.disconnectReader`, :meth:`getData`
        '''
        ok = self.dataChanged.disconnect(slot)
        self.__readerSlots.remove((weakref.ref(slot.__self__), slot.__name__))

    def isDataSet(self):
        '''Whether the data has been set at least once or if it is uninitialized

        :return: (bool) True if the data has been set. False it is uninitialized'''
        return self.__isDataSet

    def info(self):
        readers = ["%s::%s" % (repr(r()), s) for r, s in self.__readerSlots]
        writers = ["%s::%s" % (repr(r()), s) for r, s in self.__writerSignals]
        return "UID: %s\n\t Readers (%i):%s\n\t Writers (%i):%s\n" % (self.__dataUID, len(readers),
                                                                      readers, len(writers), writers)

    def readerCount(self):
        '''returns the number of currently registered readers of this model

        :return: (int)
        '''
        return len(self.__readerSlots)

    def writerCount(self):
        '''returns the number of currently registered writers of this model

        :return: (int)
        '''
        return len(self.__writerSignals)


class SharedDataManager(QtCore.QObject):
    '''
    A Factory of :class:`DataModel` objects. The :meth:`__getDataModel` method
    ensures that the created DataModels are singletons. DataModels are not kept
    alive unless there at least some Reader or Writer registered to it (or
    another object referencing them)
    '''

    def __init__(self, parent):
        QtCore.QObject.__init__(self, parent)
        self.__models = {}

    def __getDataModel(self, dataUID):
        '''
        Returns the :class:`DataModel` object for the given data UID (which is a singleton).
        If it does not previously exist, it creates one).

        .. note:: This is a private method. You are probably more interested
                  in using :meth:`connectReader` and :meth:`connectWriter`

        :param dataUID: (str) the unique identifier of the data

        :return: (DataModel)

        .. seealso:: :meth:`connectReader`,  :meth:`connectWriter`, :class:`DataModel`
        '''
        if dataUID not in self.__models:
            self.__models[dataUID] = DataModel(self, dataUID)
        return self.__models[dataUID]

    def getDataModelProxy(self, dataUID, callback=None):
        '''
        Returns a :class:`weakref.proxy` to a :class:`DataModel` object for the
        given data UID or None if the UID is not registered.

        .. note:: The underlying :class:`DataModel` object may cease to exist if
                  all its readers and writers are unregistered.

        :param dataUID: (str) the unique identifier of the data
        :param callback: (callable) same as in :class:`weakref.ref` callback parameter

        :return: (weakref.proxy or None)

        .. seealso:: :meth:`connectReader`,  :meth:`connectWriter`, :class:`DataModel`
        '''
        if dataUID not in self.__models:
            return None
        dm = self.__getDataModel(dataUID)
        return weakref.proxy(dm, callback)

    def connectReader(self, dataUID, slot, readOnConnect=True):
        '''
        Registers the given slot method to receive notifications whenever the
        data identified by dataUID is changed.

        Note that it returns the :meth:`DataModel.getData` method for the given data
        UID, which can be used for reading the data at any moment.

        :param dataUID: (str) the unique identifier of the data
        :param slot: (callable) a method that will be called when the data changes
                     this slot will be the receiver of a signal which has the
                     data as its first argument.
        :param readOnConnect: (bool) if True (default) the slot will be called
                              immediately with the current value of the data
                              if the data has been already initialized

        :return: (callable) a callable that can be used for reading the data

        .. seealso:: :meth:`connectWriter`, :meth:`__getDataModel`
        '''
        m = self.__getDataModel(dataUID)
        m.connectReader(slot, readOnConnect=True)
        if _DEBUG:
            # @todo: comment this line out. ONLY FOR DEBUGGING
            m.connectReader(self.debugReader)
        return m.getData

    def connectWriter(self, dataUID, writer, signalname):
        '''
        Registers the given writer object as a changer of the shared data
        identified by dataUID. The writer is then expected to emit a
        `QtCore.SIGNAL(signalname)` with the new data as the first parameter

        Note that it returns the :meth:`DataModel.setData` method for the given data
        UID, which can be used for changing the data at any moment.

        :param dataUID: (str) the unique identifier of the data
        :param writer: (QObject) object that will change the data
        :param signalname: (str) the signal name that will notify changes
                           of the data

        :return: (callable) a callable that can be used for setting the data.
                 When using it, one parameter has to be passed containing the
                 new data

        .. seealso:: :meth:`connectWriter`, :meth:`__getDataModel`
        '''
        m = self.__getDataModel(dataUID)
        m.connectWriter(writer, signalname)
        if _DEBUG:
            # @todo: comment this line out. ONLY FOR DEBUGGING
            m.connectReader(self.debugReader)
        return m.setData

    def disconnectWriter(self, dataUID, writer, signalname):
        '''Unregister the given object as writer of the shared data

        :param dataUID: (str) the unique identifier of the data
        :param writer: (QObject) object to unregister
        :param signalname: (str) the signal that was registered

        .. seealso:: :meth:`DataModel.disconnectWriter`
        '''
        m = self.__getDataModel(dataUID)
        m.disconnectWriter(writer, signalname)
        if m.readerCount() < 1 and m.writerCount() < 1:
            self.__models.pop(dataUID)

    def disconnectReader(self, dataUID, slot):
        '''Unregister the given method as data receiver

        :param dataUID: (str) the unique identifier of the data
        :param slot: (str) the slot that was registered

        .. seealso:: :meth:`DataModel.disconnectReader`
        '''
        m = self.__getDataModel(dataUID)
        m.disconnectReader(slot)
        if m.readerCount() < 1 and m.writerCount() < 1:
            self.__models.pop(dataUID)

    def activeDataUIDs(self):
        '''
        Returns a list of currently shared data. Note that this list only
        reflects the situation at the moment of calling this method: a given
        DataModel may die at any moment if there are no references to it.

        :returns: (list<str>) UIDs of currently shared data.
        '''
        return list(self.__models.keys())

    def debugReader(self, data):
        '''
        A slot which you can connect as a reader for debugging. It will print info to the stdout
        '''
        print("SharedDataManager: \n\tSender=: %s\n\tData=%s" % (self.sender(), repr(data)))

    def info(self):
        s = ""
        for uid, m in sorted(self.__models.items()):
            s += m.info() + '\n'
        return s



#!/usr/bin/env python

import sys
from taurus.qt.qtgui.plot import TaurusPlot
from taurus.qt.qtgui.application import TaurusApplication

app = TaurusApplication(sys.argv, cmd_line_parser=None)
##########################
# BEGUIN EXAMPLE CODE
##########################

panel = TaurusPlot()
model = ['sys/taurustest/1/abscissas|sys/taurustest/1/curve']
panel.setModel(model)

########################
# END EXAMPLE CODE
########################
panel.show()
sys.exit(app.exec_())



#!/usr/bin/env python

#############################################################################
##
# This file is part of Taurus
##
# http://taurus-scada.org
##
# Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
# Taurus is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
##
# Taurus is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
##
# You should have received a copy of the GNU Lesser General Public License
# along with Taurus.  If not, see <http://www.gnu.org/licenses/>.
##
#############################################################################

"""
Tango extension for taurus core mode.
The Tango extension implements :mod:`taurus.core` objects that connect to Tango
objects. The official scheme name is, obviously, 'tango'.

This extension maps the (Py)Tango objects into Taurus objects as follows:

- A Tango database is represented as a subclass of
  :class:`taurus.core.TaurusAuthority`
- A Tango device is represented as a subclass of
  :class:`taurus.core.TaurusDevice`
- A Tango device attribute is represented as a subclass
  :class:`taurus.core.TaurusAttribute`

You should never create objects from the above classes directly. Instead, you
should use :meth:`taurus.Authority`, :meth:`taurus.Device`,
and :meth:`taurus.Attribute` helper functions,
as in the examples below, or if you require more control, use the
:class:`taurus.core.taurusmanager.TaurusManager` or
:class:`taurus.core.taurusfactory.TaurusFactory` APIs.

Here are some examples:

The Taurus Authority associated with the Tango database
running on host "machine01" and port 10000 is named "//machine:10000"
(note that Taurus authority names are always prefixed by "//", to comply with
RFC3986). And you can get the corresponding Taurus Authority object as:

    >>> import taurus
    >>> my_db = taurus.Authority('tango://machine:10000')

If "tango" is configured as the default scheme for Taurus, the 'tango:' prefix
could be avoided and same database could be accessed as::

    >>> my_db = taurus.Authority('//machine:10000')

Now, assume that a TangoTest device is registered  in the above database as
``sys/tg_test/1``. In this case, the corresponding Taurus device full name
would be ``tango://machine:10000/sys/tg_test/1`` and it could be accessed as::

    >>> import taurus
    >>> my_device = taurus.Device('tango://machine:10000/sys/tg_test/1')

If "tango" is configured as the default scheme for Taurus, the previous name
could be shortened to ``//machine:10000/sys/tg_test/1`` or even to
``sys/tg_test/1`` if the TANGO_HOST environment variable
(or tango.rc file) point to `machine:10000` as the default tango database.
Furthermore, if, on top of that,  this device is aliased as ``tgtest1`` in the
database, it could be accessed as::

    >>> import taurus
    >>> my_device = taurus.Device('tgtest1')

Similarly, accessing the ``ampli`` attribute from the above Tango device can be
done using its full name::

    >>> import taurus
    >>> my_attr = taurus.Attribute('tango://machine:10000/sys/tg_test/1/ampli')

And of course shorter names can also be used for attributes. Following the
examples for the device above, the following names could also have been passed
to :meth:`taurus.Attribute`:

- ``//machine:10000/sys/tg_test/1/ampli``
- ``sys/tg_test/1/ampli``
- ``tgtest1/ampli``

Finally, the TangoFactory object can be accessed as::

    >>> import taurus
    >>> tg_factory = taurus.Factory('tango')


.. note:: Previous to TEP3, a RFC3986 non-compliant syntax was used for the
          Tango scheme (e.g., allowing names such as ``tango://a/b/c/d`` -note
          the double slash which should not be there).
          This syntax is now deprecated and should not be used. Taurus will
          issue warnings if detected.
"""

from __future__ import absolute_import

from .enums import *
from .tangodatabase import *
from .tangodevice import *
from .tangofactory import *
from .tangoattribute import *
from .tangoconfiguration import *


__docformat__ = "restructuredtext"



#!/usr/bin/env python

#############################################################################
##
# This file is part of Taurus
##
# http://taurus-scada.org
##
# Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
# Taurus is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
##
# Taurus is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
##
# You should have received a copy of the GNU Lesser General Public License
# along with Taurus.  If not, see <http://www.gnu.org/licenses/>.
##
#############################################################################

"""This module contains a :class:`Timer` class"""

__all__ = ["Timer"]

__docformat__ = "restructuredtext"

import time
import threading

from .log import Logger


class Timer(Logger):
    """
    Timer Object.

    Interval in seconds (The argument may be a floating point number for
    subsecond precision).
    If strict_timing is True, the timer will try to compensate for drifting
    due to the time it takes to execute function in each loop.
    """

    def __init__(self, interval, function, parent, strict_timing=True,
                 *args, **kwargs):
        Logger.__init__(self, 'Timer on ' + function.__name__, parent)
        self.__lock = threading.Lock()
        self.__interval = interval
        self.__function = function
        self.__args = args
        self.__kwargs = kwargs
        self.__loop = False
        self.__alive = False
        self.__start_nb = 0
        self.__thread = None
        self.__strict_timing = strict_timing

    def start(self):
        """ Start Timer Object """
        self.__lock.acquire()
        try:
            if not self.__alive:
                self.debug("Timer::start()")
                self.__loop = True
                self.__alive = True
                self.__start_nb += 1
                thread_name = "TimerLoop %d" % self.__start_nb
                self.__thread = threading.Thread(
                    target=self.__run, name=thread_name)
                self.__thread.setDaemon(True)
                self.__thread.start()
        finally:
            self.__lock.release()

    def stop(self, sync=False):
        """ Stop Timer Object """
        self.debug("Timer::stop()")
        self.__lock.acquire()
        self.__loop = False
        self.__lock.release()
        if sync and self.__thread is not None:
            self.__thread.join()

    def __run(self):
        """ Private Thread Function """
        self.debug("Timer thread starting")
        next_time = time.time() + self.__interval
        while self.__loop:
            self.__function(*self.__args, **self.__kwargs)
            nap = self.__interval
            if self.__strict_timing:
                curr_time = time.time()
                nap = max(0, next_time - curr_time)
                if curr_time > next_time:
                    self.warning("loop function took more than loop interval (%ss)",
                                 self.__interval)
            next_time += self.__interval
            time.sleep(nap)
        self.__alive = False
        self.debug("Timer thread ending")



#!/usr/bin/env python

#############################################################################
##
# This file is part of Taurus
##
# http://taurus-scada.org
##
# Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
# Taurus is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
##
# Taurus is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
##
# You should have received a copy of the GNU Lesser General Public License
# along with Taurus.  If not, see <http://www.gnu.org/licenses/>.
##
#############################################################################

"""This module provides a base widget that can be used to display a taurus
model in a tree widget"""

from __future__ import absolute_import

from taurus.qt.qtgui.model import TaurusBaseModelWidget
from .qtree import QBaseTreeWidget


__all__ = ["TaurusBaseTreeWidget"]

__docformat__ = 'restructuredtext'


class TaurusBaseTreeWidget(QBaseTreeWidget, TaurusBaseModelWidget):

    def __init__(self, parent=None, designMode=False, with_navigation_bar=True,
                 with_filter_widget=True, perspective=None, proxy=None):
        self.call__init__(QBaseTreeWidget, parent, designMode=designMode,
                          with_navigation_bar=with_navigation_bar,
                          with_filter_widget=with_filter_widget,
                          perspective=perspective, proxy=proxy)
        self.call__init__(TaurusBaseModelWidget, designMode=designMode)

    #-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-
    # TaurusBaseWidget overwriting
    #-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-

    def updateStyle(self):
        """overwritten from class:`taurus.qt.qtgui.base.TaurusBaseWidget`. It is called when
        the taurus model changes."""
        self.resizeColumns()



#!/usr/bin/env python

#############################################################################
##
# This file is part of Taurus
##
# http://taurus-scada.org
##
# Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
# Taurus is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
##
# Taurus is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
##
# You should have received a copy of the GNU Lesser General Public License
# along with Taurus.  If not, see <http://www.gnu.org/licenses/>.
##
#############################################################################

"""This module provides basic python dictionary/list editor widgets"""
from __future__ import print_function

import sys
from future.utils import string_types

from taurus.core.util.containers import SortedDict
from taurus.external.qt import Qt
from taurus.qt.qtgui.container import TaurusBaseContainer, TaurusWidget
from taurus.qt.qtcore.util.properties import join, djoin


__all__ = ["QDictionaryEditor", "QListEditor"]

__docformat__ = 'restructuredtext'

###############################################################################
# Methods borrowed from fandango modules


def isString(seq):
    if isinstance(seq, string_types):
        return True  # It matches most python str-like classes
    if any(s in str(type(seq)).lower() for s in ('vector', 'array', 'list',)):
        return False
    if 'qstring' == str(type(seq)).lower():
        return True  # It matches QString
    return False


def isSequence(seq, INCLUDE_GENERATORS=True):
    """ It excludes Strings, dictionaries but includes generators"""
    if any(isinstance(seq, t) for t in (list, set, tuple)):
        return True
    if isString(seq):
        return False
    if hasattr(seq, 'items'):
        return False
    if INCLUDE_GENERATORS:
        if hasattr(seq, '__iter__'):
            return True
    elif hasattr(seq, '__len__'):
        return True
    return False


def isDictionary(seq):
    """ It includes dicts and also nested lists """
    if isinstance(seq, dict):
        return True
    if hasattr(seq, 'items') or hasattr(seq, 'iteritems'):
        return True
    if seq and isSequence(seq) and isSequence(seq[0]):
        if seq[0] and not isSequence(seq[0][0]):
            return True  # First element of tuple must be hashable
    return False


def dict2array(dct):
    """ Converts a dictionary in a table of data, lists are unnested columns """
    data, table = {}, []
    data['nrows'], data['ncols'] = 0, 2 if isDictionary(dct) else 1

    def expand(d, level):  # ,nrows=nrows,ncols=ncols):
        # self.debug('\texpand(%s(%s),%s)'%(type(d),d,level))
        items = list(d.items()) if isinstance(d, SortedDict) else sorted(
            list(d.items()) if hasattr(d, 'items') else d)
        for k, v in items:
            zero = data['nrows']
            data[(data['nrows'], level)] = k
            if isDictionary(v):
                data['ncols'] += 1
                expand(v, level + 1)
            else:
                if not isSequence(v):
                    v = [v]
                for t in v:
                    data[(data['nrows'], level + 1)] = t
                    data['nrows'] += 1
            #for i in range(zero+1,nrows): data[(i,level)] = None
    expand(dct, 0)
    [table.append([]) for r in range(data.pop('nrows'))]
    [table[r].append(None) for c in range(data.pop('ncols'))
     for r in range(len(table))]
    for coord, value in data.items():
        table[coord[0]][coord[1]] = value
    return table


def array2dict(table):
    """ Converts a table in a dictionary of left-to-right nested date, unnested columns are lists"""
    nrows, ncols = len(table), len(table[0])
    r, c = 0, 0

    def expand(r, c, end):
        print('expand(%s,%s,%s)' % (r, c, end))
        i0, t0 = r, table[r][c]
        if not t0:
            return t0
        if c + 1 < ncols and (table[r][c + 1] or not c):
            d = {}
            keys = []
            new_end = r + 1
            for i in range(r + 1, end + 1):
                t = table[i][c] if i < end else None
                if t or i >= end:
                    # start,name,stop for each key
                    keys.append((i0, t0, new_end))
                    t0, i0 = t, i
                new_end = i + 1
            for i, key, new_end in keys:
                nd = expand(i, c + 1, new_end)
                d[key] = nd if key not in d else djoin(d.get(key), nd)
            print('expand(%s to %s,%s): %s' % (r, end, c, d))
            return d
        else:
            d = [table[i][c] for i in range(r, end)]
            print('expand(%s to %s,%s): %s' % (r, end, c, d))
            return d
    data = expand(0, 0, nrows)
    return data

###############################################################################


class QBaseDictionaryEditor(Qt.QDialog, TaurusBaseContainer):

    def __init__(self, parent=None, designMode=None, title=None):
        self.data = {}  # An {(x,y):value} array
        self.title = title
        self.dctmodel = SortedDict()
        self.callback = None
        self.call__init__wo_kw(Qt.QDialog, parent)
        self.call__init__(TaurusBaseContainer, type(
            self).__name__, designMode=designMode)  # defineStyle called from here

    @classmethod
    def main(klass, args=None, title='', modal=False, callback=None):
        dialog = klass()
        dialog.setModal(modal)
        dialog.setCallback(callback)
        dialog.setModifiableByUser(True)
        dialog.setWindowTitle(title or klass.title or klass.__name__)
        if args:
            dialog.setModel(args)  # [0] if isSequence(args) else args)
        dialog.show()
        return dialog

    def defineStyle(self):
        self.info('QBaseDictionaryEditor.defineStyle()')
        # self.setWindowTitle('DictionaryEditor')
        self.label = Qt.QLabel(
            'Dictionary as a nested tree: {key1:[val1,val2],key2:[val3]')
        #self.value = Qt.QLabel()
        self.table = Qt.QTableWidget()  # TaurusBaseTable()
        self.table.horizontalHeader().setStretchLastSection(True)
        # self.table.verticalHeader().setStretchLastSection(True)
        self.baccept = Qt.QPushButton('Apply')
        self.bcancel = Qt.QPushButton('Cancel')
        self.baddColumn = Qt.QPushButton()
        self.baddRow = Qt.QPushButton()
        [(b.setFixedSize(Qt.QSize(20, 20)), b.setIcon(Qt.Qicon(
            'designer:plus.png'))) for b in (self.baddColumn, self.baddRow)]
        self.setLayout(Qt.QGridLayout())
        self.layout().addWidget(self.label, 0, 0, 1, 5)
        # self.layout().addWidget(self.value,1,0,1,3)
        self.layout().addWidget(self.baddColumn, 2, 5, 1, 1)
        self.layout().addWidget(self.table, 2, 0, 5, 5)
        self.layout().addWidget(self.baddRow, 7, 0, 1, 1)
        self.layout().addWidget(self.baccept, 8, 3, 1, 1)
        self.layout().addWidget(self.bcancel, 8, 4, 1, 1)
        self.baccept.clicked.connect(self.save)
        self.bcancel.clicked.connect(self.close)
        self.baddRow.clicked.connect(self.addRow)
        self.baddColumn.clicked.connect(self.addColumn)
        self.reject.connect(self.close)

    def addRow(self):
        self.table.setRowCount(self.table.rowCount() + 1)
        self.table.resizeRowsToContents()
        self.table.update()

    def addColumn(self):
        self.table.horizontalHeader().setStretchLastSection(False)
        self.table.setColumnCount(self.table.columnCount() + 1)
        self.table.resizeColumnsToContents()
        self.table.horizontalHeader().setStretchLastSection(True)
        self.table.update()

    def setCallback(self, callback):
        self.callback = callback

    def getCellText(self, row, column):
        if row >= self.table.rowCount() or column >= self.table.columnCount():
            v = None
        else:
            i = self.table.item(row, column)
            if i is None:
                v = i
            else:
                v = str(i.text()).strip()
        self.debug('getCellText(%s,%s): %s' % (row, column, v))
        return v

    def setCellText(self, row, column, value, bold=False, italic=False):
        i = self.table.item(row, column) or Qt.QTableWidgetItem()
        i.setText(str(value if value is not None else ''))
        if bold or italic:
            f = i.font()
            if bold:
                f.setBold(True)
            if italic:
                f.setItalic(True)
            i.setFont(f)
        self.table.setItem(row, column, i)

    def setModel(self, model):
        raise Exception('setModel(self,model)!')

    def updateStyle(self):
        raise Exception('updateStyle(self)!')

    def getValues(self):
        raise Exception('getValues(self)!')

    def save(self):
        raise Exception('save(self)!')

###############################################################################


class QDictionaryEditor(QBaseDictionaryEditor):

    def setModel(self, model):
        self.info('DictionaryEditor.setModel(%s(%s))' % (type(model), model))
        self.dctmodel = eval(model) if isString(model) else model
        # self.updateStyle() called from the property setter
        TaurusBaseContainer.setModel(self, model)

    def getModelClass(self):
        return dict

    def updateStyle(self):
        # self.value.setText(str(self.dctmodel))
        data = dict2array(self.dctmodel)
        self.nrows, self.ncols = len(data), len(data[0])
        self.info(data)
        self.table.setRowCount(self.nrows)
        self.table.setColumnCount(self.ncols)
        for r in range(self.nrows):
            for c in range(self.ncols):
                self.setCellText(r, c, data[r][c],
                                 bold=(not c), italic=(c == 1))
        self.table.resizeRowsToContents()
        self.table.resizeColumnsToContents()
        self.update()

    def getValues(self):
        nrows, ncols = self.table.rowCount(), self.table.columnCount()
        table = [[self.getCellText(r, c) for c in range(ncols)]
                 for r in range(nrows)]
        self.data = array2dict(table)  # It returns a SortedDict
        self.info('getValues(): %s' % str(self.data))
        return self.data

    def save(self):
        self.getValues()
        self.info('DictionaryEditor.save(): %s' % self.data)
        if self.callback:
            self.callback(self.data)
        elif self.dctmodel is None:
            self.dctmodel = self.data
        else:  # Overwriting dctmodel
            self.dctmodel.clear()
            self.dctmodel.update(self.data)
        if self.callback:
            self.callback(self.data)  # A SortedDict is passed here
        self.updateStyle()

###############################################################################


class QListEditor(QBaseDictionaryEditor):

    def defineStyle(self):
        QBaseDictionaryEditor.defineStyle(self)
        self.table.setColumnCount(1)
        self.baddColumn.hide()

    def setModel(self, model):
        self.info('DictionaryEditor.setModel(%s(%s))' % (type(model), model))
        if isString(model):
            try:
                self.dctmodel = list(eval(model)) if any(
                    c in model for c in ('{', '[', '(')) else [model]
            except:
                self.dctmodel = [model]
        else:
            self.dctmodel = model
        # self.updateStyle() called from the property setter
        TaurusBaseContainer.setModel(self, model)

    def getModelClass(self):
        return list

    def updateStyle(self):
        # self.value.setText(str(self.dctmodel))
        data = list(self.dctmodel)
        self.nrows, self.ncols = len(data), 1
        self.table.setRowCount(self.nrows)
        self.table.setColumnCount(self.ncols)
        for r in range(self.nrows):
            self.setCellText(r, 0, data[r])
        self.table.resizeRowsToContents()
        self.table.horizontalHeader().setStretchLastSection(True)
        self.update()

    def getValues(self):
        nrows = self.table.rowCount()
        self.data = [self.getCellText(r, 0)for r in range(nrows)]
        return self.data

    def save(self):
        self.getValues()
        self.info('DictionaryEditor.save(%s(%s)): %s' %
                  (type(self.data), self.data, self.callback))
        if self.callback:
            self.callback(self.data)
        elif self.dctmodel is None:
            self.dctmodel = self.data
        else:  # Overwriting dctmodel
            if isSequence(self.dctmodel):
                while len(self.dctmodel):
                    self.dctmodel.pop(0)
                self.dctmodel.extend(self.data)
            elif isDictionary(self.dctmodel):
                self.dctmodel.clear()
                self.dctmodel = djoin(self.dctmodel, self.data)
        return self.data


###############################################################################

def prepare():
    from taurus.qt.qtgui.application import TaurusApplication
    app = TaurusApplication(app_name='DictionaryEditor')
    args = app.get_command_line_args()
    return app, args

if __name__ == '__main__':
    app, args = prepare()
    dialog = QDictionaryEditor.main(args)
    sys.exit(app.exec_())
